"""
Functions for handling collisions with geometry.

These are the pure python versions of the functions.

Point in polygon function from https://www.geeksforgeeks.org/how-to-check-if-a-given-point-lies-inside-a-polygon/
"""

from sys import maxsize as sys_int_maxsize

from arcade.types import Point2, Point2List


def are_polygons_intersecting(poly_a: Point2List, poly_b: Point2List) -> bool:
    """
    Check if two polygons intersect.

    Args:
        poly_a: List of points that define the first polygon.
        poly_b: List of points that define the second polygon.

    Returns:
        ``True`` if polygons intersect, ``False`` otherwise
    """
    # if either are [], they don't intersect
    if not poly_a or not poly_b:
        return False
    for polygon in (poly_a, poly_b):
        for i1 in range(len(polygon)):
            i2 = (i1 + 1) % len(polygon)
            projection_1 = polygon[i1]
            projection_2 = polygon[i2]

            normal = (
                projection_2[1] - projection_1[1],
                projection_1[0] - projection_2[0],
            )

            min_a, min_b = (float("inf"),) * 2
            max_a, max_b = (-float("inf"),) * 2

            for poly in poly_a:
                projected = normal[0] * poly[0] + normal[1] * poly[1]

                if projected < min_a:
                    min_a = projected
                if projected > max_a:
                    max_a = projected

            for poly in poly_b:
                projected = normal[0] * poly[0] + normal[1] * poly[1]

                if projected < min_b:
                    min_b = projected
                if projected > max_b:
                    max_b = projected

            # Avoid typing.cast() because this is a very hot path
            if max_a <= min_b or max_b <= min_a:
                return False

    return True


def is_point_in_box(p: Point2, q: Point2, r: Point2) -> bool:
    """
    Checks if point ``q`` is inside the box defined by ``p`` and ``r``.

    Args:
        p (Point2): Start of box
        q (Point2): Point to check
        r (Point2): End of box

    Returns:
        ``True`` or ``False`` depending if point is in the box.
    """
    return (
        (q[0] <= max(p[0], r[0]))
        and (q[0] >= min(p[0], r[0]))
        and (q[1] <= max(p[1], r[1]))
        and (q[1] >= min(p[1], r[1]))
    )


def get_triangle_orientation(p: Point2, q: Point2, r: Point2) -> int:
    """
    Find the orientation of a triangle defined by (p, q, r)

    The function returns the following integer values:

      * 0 --> p, q, and r are collinear
      * 1 --> Clockwise
      * 2 --> Counterclockwise

    Args:
        p: Point 1
        q: Point 2
        r: Point 3

    Returns:
        int: 0, 1, or 2 depending on the orientation
    """
    val = ((q[1] - p[1]) * (r[0] - q[0])) - ((q[0] - p[0]) * (r[1] - q[1]))

    if val == 0:
        return 0  # collinear
    if val > 0:
        return 1  # clockwise
    else:
        return 2  # counter-clockwise


def are_lines_intersecting(p1: Point2, q1: Point2, p2: Point2, q2: Point2) -> bool:
    """
    Given two lines defined by points p1, q1 and p2, q2, the function
    returns true if the two lines intersect.

    Args:
        p1: Point 1
        q1: Point 2
        p2: Point 3
        q2: Point 4

    Returns:
        bool: ``True`` or ``False`` depending if lines intersect
    """
    o1 = get_triangle_orientation(p1, q1, p2)
    o2 = get_triangle_orientation(p1, q1, q2)
    o3 = get_triangle_orientation(p2, q2, p1)
    o4 = get_triangle_orientation(p2, q2, q1)

    # General case
    if (o1 != o2) and (o3 != o4):
        return True

    # Special Cases
    # p1, q1 and p2 are collinear and p2 lies on segment p1q1
    if (o1 == 0) and is_point_in_box(p1, p2, q1):
        return True

    # p1, q1 and p2 are collinear and q2 lies on segment p1q1
    if (o2 == 0) and is_point_in_box(p1, q2, q1):
        return True

    # p2, q2 and p1 are collinear and p1 lies on segment p2q2
    if (o3 == 0) and is_point_in_box(p2, p1, q2):
        return True

    # p2, q2 and q1 are collinear and q1 lies on segment p2q2
    if (o4 == 0) and is_point_in_box(p2, q1, q2):
        return True

    return False


def is_point_in_polygon(x: float, y: float, polygon: Point2List) -> bool:
    """
    Checks if a point is inside a polygon of three or more points.

    Args:
        x: X coordinate of point
        y: Y coordinate of point
        polygon: List of points that define the polygon.

    Returns:
        bool: ``True`` or ``False`` depending if point is inside polygon
    """
    p = x, y
    n = len(polygon)

    # There must be at least 3 vertices
    # in polygon
    if n < 3:
        return False

    # Create a point for line segment
    # from p to infinite
    extreme = (sys_int_maxsize, p[1])

    # To count number of points in polygon
    # whose y-coordinate is equal to
    # y-coordinate of the point
    decrease = 0
    count = i = 0

    while True:
        next_item = (i + 1) % n

        if polygon[i][1] == p[1]:
            decrease += 1

        # Check if the line segment from 'p' to
        # 'extreme' intersects with the line
        # segment from 'polygon[i]' to 'polygon[next]'
        if are_lines_intersecting(polygon[i], polygon[next_item], p, extreme):
            # If the point 'p' is collinear with line
            # segment 'i-next', then check if it lies
            # on segment. If it lies, return true, otherwise false
            if get_triangle_orientation(polygon[i], p, polygon[next_item]) == 0:
                return is_point_in_box(
                    polygon[i],
                    p,
                    polygon[next_item],
                )

            count += 1

        i = next_item

        if i == 0:
            break

    # Reduce the count by decrease amount
    # as these points would have been added twice
    count -= decrease

    # Return true if count is odd, false otherwise
    return count % 2 == 1



from __future__ import annotations

import weakref
from ctypes import byref, c_void_p
from typing import TYPE_CHECKING, Sequence

from pyglet import gl

from .buffer import Buffer
from .program import Program
from .types import BufferDescription, GLenumLike, GLuintLike, gl_name

if TYPE_CHECKING:
    from arcade.gl import Context

# Index buffer types based on index element size
index_types = [
    None,  # 0 (not supported)
    gl.GL_UNSIGNED_BYTE,  # 1 ubyte8
    gl.GL_UNSIGNED_SHORT,  # 2 ubyte16
    None,  # 3 (not supported)
    gl.GL_UNSIGNED_INT,  # 4 ubyte32
]


class VertexArray:
    """
    Wrapper for Vertex Array Objects (VAOs).

    This objects should not be instantiated from user code.
    Use :py:class:`arcade.gl.Geometry` instead. It will create VAO instances for you
    automatically. There is a lot of complex interaction between programs
    and vertex arrays that will be done for you automatically.

    Args:
        ctx:
            The context this object belongs to
        program:
            The program to use
        content:
            List of BufferDescriptions
        index_buffer:
            Index/element buffer
        index_element_size:
            Byte size of the index buffer datatype.
    """

    __slots__ = (
        "_ctx",
        "glo",
        "_program",
        "_content",
        "_ibo",
        "_index_element_size",
        "_index_element_type",
        "_num_vertices",
        "__weakref__",
    )

    def __init__(
        self,
        ctx: Context,
        program: Program,
        content: Sequence[BufferDescription],
        index_buffer: Buffer | None = None,
        index_element_size: int = 4,
    ) -> None:
        self._ctx = ctx
        self._program = program
        self._content = content

        self.glo = glo = gl.GLuint()
        """The OpenGL resource ID"""

        self._num_vertices = -1
        self._ibo = index_buffer
        self._index_element_size = index_element_size
        self._index_element_type = index_types[index_element_size]

        self._build(program, content, index_buffer)

        if self._ctx.gc_mode == "auto":
            weakref.finalize(self, VertexArray.delete_glo, self.ctx, glo)

        self.ctx.stats.incr("vertex_array")

    def __repr__(self) -> str:
        return f"<VertexArray {self.glo.value}>"

    def __del__(self) -> None:
        # Intercept garbage collection if we are using Context.gc()
        if self._ctx.gc_mode == "context_gc" and self.glo.value > 0:
            self._ctx.objects.append(self)

    @property
    def ctx(self) -> Context:
        """The Context this object belongs to."""
        return self._ctx

    @property
    def program(self) -> Program:
        """The assigned program."""
        return self._program

    @property
    def ibo(self) -> Buffer | None:
        """Element/index buffer."""
        return self._ibo

    @property
    def num_vertices(self) -> int:
        """The number of vertices."""
        return self._num_vertices

    def delete(self) -> None:
        """
        Destroy the underlying OpenGL resource.

        Don't use this unless you know exactly what you are doing.
        """
        VertexArray.delete_glo(self._ctx, self.glo)
        self.glo.value = 0

    @staticmethod
    def delete_glo(ctx: Context, glo: gl.GLuint) -> None:
        """
        Delete the OpenGL resource.

        This is automatically called when this object is garbage collected.
        """
        # If we have no context, then we are shutting down, so skip this
        if gl.current_context is None:
            return

        if glo.value != 0:
            gl.glDeleteVertexArrays(1, byref(glo))
            glo.value = 0

        ctx.stats.decr("vertex_array")

    def _build(
        self, program: Program, content: Sequence[BufferDescription], index_buffer: Buffer | None
    ) -> None:
        """
        Build a vertex array compatible with the program passed in.

        This method will bind the vertex array and set up all the vertex attributes
        according to the program's attribute specifications.

        Args:
            program:
                The program to use
            content:
                List of BufferDescriptions
            index_buffer:
                Index/element buffer
        """
        gl.glGenVertexArrays(1, byref(self.glo))
        gl.glBindVertexArray(self.glo)

        if index_buffer is not None:
            gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, index_buffer.glo)

        # Lookup dict for BufferDescription attrib names
        descr_attribs = {attr.name: (descr, attr) for descr in content for attr in descr.formats}

        # Build the vao according to the shader's attribute specifications
        for _, prog_attr in enumerate(program.attributes):
            # Do we actually have an attribute with this name in buffer descriptions?
            if prog_attr.name is not None and prog_attr.name.startswith("gl_"):
                continue
            try:
                buff_descr, attr_descr = descr_attribs[prog_attr.name]
            except KeyError:
                raise ValueError(
                    (
                        f"Program needs attribute '{prog_attr.name}', but is not present in buffer "
                        f"description. Buffer descriptions: {content}"
                    )
                )

            # Make sure components described in BufferDescription and in the shader match
            if prog_attr.components != attr_descr.components:
                raise ValueError(
                    (
                        f"Program attribute '{prog_attr.name}' has {prog_attr.components} "
                        f"components while the buffer description has {attr_descr.components} "
                        " components. "
                    )
                )

            gl.glEnableVertexAttribArray(prog_attr.location)
            gl.glBindBuffer(gl.GL_ARRAY_BUFFER, buff_descr.buffer.glo)

            # TODO: Detect normalization
            normalized = gl.GL_TRUE if attr_descr.name in buff_descr.normalized else gl.GL_FALSE

            # Map attributes groups
            float_types = (gl.GL_FLOAT, gl.GL_HALF_FLOAT)
            double_types = (gl.GL_DOUBLE,)
            int_types = (
                gl.GL_INT,
                gl.GL_UNSIGNED_INT,
                gl.GL_SHORT,
                gl.GL_UNSIGNED_SHORT,
                gl.GL_BYTE,
                gl.GL_UNSIGNED_BYTE,
            )
            attrib_type = attr_descr.gl_type
            # Normalized integers must be mapped as floats
            if attrib_type in int_types and buff_descr.normalized:
                attrib_type = prog_attr.gl_type

            # Sanity check attribute types between shader and buffer description
            if attrib_type != prog_attr.gl_type:
                raise ValueError(
                    (
                        f"Program attribute '{prog_attr.name}' has type "
                        f"{gl_name(prog_attr.gl_type)} "
                        f"while the buffer description has type {gl_name(attr_descr.gl_type)}. "
                    )
                )

            if attrib_type in float_types:
                gl.glVertexAttribPointer(
                    prog_attr.location,  # attrib location
                    attr_descr.components,  # 1, 2, 3 or 4
                    attr_descr.gl_type,  # GL_FLOAT etc
                    normalized,  # normalize
                    buff_descr.stride,
                    c_void_p(attr_descr.offset),
                )
            elif attrib_type in double_types:
                gl.glVertexAttribLPointer(
                    prog_attr.location,  # attrib location
                    attr_descr.components,  # 1, 2, 3 or 4
                    attr_descr.gl_type,  # GL_DOUBLE etc
                    buff_descr.stride,
                    c_void_p(attr_descr.offset),
                )
            elif attrib_type in int_types:
                gl.glVertexAttribIPointer(
                    prog_attr.location,  # attrib location
                    attr_descr.components,  # 1, 2, 3 or 4
                    attr_descr.gl_type,  # GL_FLOAT etc
                    buff_descr.stride,
                    c_void_p(attr_descr.offset),
                )
            else:
                raise ValueError(f"Unsupported attribute type: {attr_descr.gl_type}")

            # print((
            #     f"gl.glVertexAttribXPointer(\n"
            #     f"    {prog_attr.location},  # attrib location\n"
            #     f"    {attr_descr.components},  # 1, 2, 3 or 4\n"
            #     f"    {attr_descr.gl_type},  # GL_FLOAT etc\n"
            #     f"    {normalized},  # normalize\n"
            #     f"    {buff_descr.stride},\n"
            #     f"    c_void_p({attr_descr.offset}),\n"
            # ))
            # TODO: Sanity check this
            if buff_descr.instanced:
                gl.glVertexAttribDivisor(prog_attr.location, 1)

    def render(
        self, mode: GLenumLike, first: int = 0, vertices: int = 0, instances: int = 1
    ) -> None:
        """
        Render the VertexArray to the currently active framebuffer.

        Args:
            mode:
                Primitive type to render. TRIANGLES, LINES etc.
            first:
                The first vertex to render from
            vertices:
                Number of vertices to render
            instances:
                OpenGL instance, used in using vertices over and over
        """
        gl.glBindVertexArray(self.glo)
        if self._ibo is not None:
            # HACK: re-bind index buffer just in case.
            #       pyglet rendering was somehow replacing the index buffer.
            gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, self._ibo.glo)
            gl.glDrawElementsInstanced(
                mode,
                vertices,
                self._index_element_type,
                first * self._index_element_size,
                instances,
            )
        else:
            gl.glDrawArraysInstanced(mode, first, vertices, instances)

    def render_indirect(self, buffer: Buffer, mode: GLuintLike, count, first, stride) -> None:
        """
        Render the VertexArray to the framebuffer using indirect rendering.

        .. Warning:: This requires OpenGL 4.3

        Args:
            buffer:
                The buffer containing one or multiple draw parameters
            mode:
                Primitive type to render. TRIANGLES, LINES etc.
            count:
                The number if indirect draw calls to run
            first:
                The first indirect draw call to start on
            stride:
                The byte stride of the draw command buffer.
                Keep the default (0) if the buffer is tightly packed.
        """
        # The default buffer stride for array and indexed
        _stride = 20 if self._ibo is not None else 16
        stride = stride or _stride
        if stride % 4 != 0 or stride < 0:
            raise ValueError(f"stride must be positive integer in multiples of 4, not {stride}.")

        # The maximum number of draw calls in the buffer
        max_commands = buffer.size // stride
        if count < 0:
            count = max_commands
        elif (first + count) > max_commands:
            raise ValueError(
                "Attempt to issue rendering commands outside the buffer. "
                f"first = {first}, count = {count} is reaching past "
                f"the buffer end. The buffer have room for {max_commands} "
                f"draw commands. byte size {buffer.size}, stride {stride}."
            )

        gl.glBindVertexArray(self.glo)
        gl.glBindBuffer(gl.GL_DRAW_INDIRECT_BUFFER, buffer._glo)
        if self._ibo:
            gl.glMultiDrawElementsIndirect(
                mode, self._index_element_type, first * stride, count, stride
            )
        else:
            gl.glMultiDrawArraysIndirect(mode, first * stride, count, stride)

    def transform_interleaved(
        self,
        buffer: Buffer,
        mode: GLenumLike,
        output_mode: GLenumLike,
        first: int = 0,
        vertices: int = 0,
        instances: int = 1,
        buffer_offset=0,
    ) -> None:
        """
        Run a transform feedback.

        Args:
            buffer:
                The buffer to write the output
            mode:
                The input primitive mode
            output_mode:
                The output primitive mode
            first:
                Offset start vertex
            vertices:
                Number of vertices to render
            instances:
                Number of instances to render
            buffer_offset:
                Byte offset for the buffer (target)
        """
        if vertices < 0:
            raise ValueError(f"Cannot determine the number of vertices: {vertices}")

        if buffer_offset >= buffer.size:
            raise ValueError("buffer_offset at end or past the buffer size")

        gl.glBindVertexArray(self.glo)
        gl.glEnable(gl.GL_RASTERIZER_DISCARD)

        if buffer_offset > 0:
            gl.glBindBufferRange(
                gl.GL_TRANSFORM_FEEDBACK_BUFFER,
                0,
                buffer.glo,
                buffer_offset,
                buffer.size - buffer_offset,
            )
        else:
            gl.glBindBufferBase(gl.GL_TRANSFORM_FEEDBACK_BUFFER, 0, buffer.glo)

        gl.glBeginTransformFeedback(output_mode)

        if self._ibo is not None:
            count = self._ibo.size // 4
            # TODO: Support first argument by offsetting pointer (second last arg)
            gl.glDrawElementsInstanced(mode, vertices or count, gl.GL_UNSIGNED_INT, None, instances)
        else:
            # print(f"glDrawArraysInstanced({mode}, {first}, {vertices}, {instances})")
            gl.glDrawArraysInstanced(mode, first, vertices, instances)

        gl.glEndTransformFeedback()
        gl.glDisable(gl.GL_RASTERIZER_DISCARD)

    def transform_separate(
        self,
        buffers: list[Buffer],
        mode: GLenumLike,
        output_mode: GLenumLike,
        first: int = 0,
        vertices: int = 0,
        instances: int = 1,
        buffer_offset=0,
    ) -> None:
        """
        Run a transform feedback writing to separate buffers.

        Args:
            buffers:
                The buffers to write the output
            mode:
                The input primitive mode
            output_mode:
                The output primitive mode
            first:
                Offset start vertex
            vertices:
                Number of vertices to render
            instances:
                Number of instances to render
            buffer_offset:
                Byte offset for the buffer (target)
        """
        if vertices < 0:
            raise ValueError(f"Cannot determine the number of vertices: {vertices}")

        # Get size from the smallest buffer
        size = min(buf.size for buf in buffers)
        if buffer_offset >= size:
            raise ValueError("buffer_offset at end or past the buffer size")

        gl.glBindVertexArray(self.glo)
        gl.glEnable(gl.GL_RASTERIZER_DISCARD)

        if buffer_offset > 0:
            for index, buffer in enumerate(buffers):
                gl.glBindBufferRange(
                    gl.GL_TRANSFORM_FEEDBACK_BUFFER,
                    index,
                    buffer.glo,
                    buffer_offset,
                    buffer.size - buffer_offset,
                )
        else:
            for index, buffer in enumerate(buffers):
                gl.glBindBufferBase(gl.GL_TRANSFORM_FEEDBACK_BUFFER, index, buffer.glo)

        gl.glBeginTransformFeedback(output_mode)

        if self._ibo is not None:
            count = self._ibo.size // 4
            # TODO: Support first argument by offsetting pointer (second last arg)
            gl.glDrawElementsInstanced(mode, vertices or count, gl.GL_UNSIGNED_INT, None, instances)
        else:
            # print(f"glDrawArraysInstanced({mode}, {first}, {vertices}, {instances})")
            gl.glDrawArraysInstanced(mode, first, vertices, instances)

        gl.glEndTransformFeedback()
        gl.glDisable(gl.GL_RASTERIZER_DISCARD)


class Geometry:
    """A higher level abstraction of the VertexArray.

    It generates VertexArray instances on the fly internally matching the incoming
    program. This means we can render the same geometry with different programs
    as long as the :py:class:`~arcade.gl.Program` and :py:class:`~arcade.gl.BufferDescription`
    have compatible attributes. This is an extremely powerful concept that allows
    for very flexible rendering pipelines and saves the user from a lot of manual
    bookkeeping.

    Geometry objects should be created through :py:meth:`arcade.gl.Context.geometry`

    Args:
        ctx:
            The context this object belongs to
        content:
            List of BufferDescriptions
        index_buffer:
            Index/element buffer
        mode:
            The default draw mode
        index_element_size:
            Byte size of the index buffer datatype.
            Can be 1, 2 or 4 (8, 16 or 32bit integer)
    """

    __slots__ = (
        "_ctx",
        "_content",
        "_index_buffer",
        "_index_element_size",
        "_mode",
        "_vao_cache",
        "_num_vertices",
        "__weakref__",
    )

    def __init__(
        self,
        ctx: "Context",
        content: Sequence[BufferDescription] | None,
        index_buffer: Buffer | None = None,
        mode: int | None = None,
        index_element_size: int = 4,
    ) -> None:
        self._ctx = ctx
        self._content = list(content or [])
        self._index_buffer = index_buffer
        self._index_element_size = index_element_size
        self._mode = mode if mode is not None else ctx.TRIANGLES
        self._vao_cache: dict[str, VertexArray] = {}
        self._num_vertices: int = -1

        if self._index_buffer and self._index_element_size not in (1, 2, 4):
            raise ValueError("index_element_size must be 1, 2, or 4")

        if content:
            # Calculate vertices. Use the minimum for now
            if self._index_buffer:
                self._num_vertices = self._index_buffer.size // self._index_element_size
            else:
                self._num_vertices = content[0].num_vertices
                for descr in self._content:
                    if descr.instanced:
                        continue
                    self._num_vertices = min(self._num_vertices, descr.num_vertices)

        # No cleanup is needed, but we want to count them
        weakref.finalize(self, Geometry._release, self._ctx)
        self._ctx.stats.incr("geometry")

    @property
    def ctx(self) -> "Context":
        """The context this geometry belongs to."""
        return self._ctx

    @property
    def index_buffer(self) -> Buffer | None:
        """Index/element buffer if supplied at creation."""
        return self._index_buffer

    @property
    def num_vertices(self) -> int:
        """
        Get or set the number of vertices.

        Be careful when modifying this properly
        and be absolutely sure what you are doing.
        """
        # TODO: Calculate this better...
        return self._num_vertices

    @num_vertices.setter
    def num_vertices(self, value: int):
        self._num_vertices = value

    def append_buffer_description(self, descr: BufferDescription):
        """
        Append a new BufferDescription to the existing Geometry.

        .. Warning:: Geometry cannot contain two BufferDescriptions which share an attribute name.
        """
        for other_descr in self._content:
            if other_descr == descr:
                raise ValueError(
                    "Geometry cannot contain two BufferDescriptions which share an "
                    f"attribute name. Found a conflict in {descr} and {other_descr}"
                )
        self._content.append(descr)

    def instance(self, program: Program) -> VertexArray:
        """
        Get the :py:class:`arcade.gl.VertexArray` compatible with this program.
        """
        vao = self._vao_cache.get(program.attribute_key)
        if vao:
            return vao

        return self._generate_vao(program)

    def render(
        self,
        program: Program,
        *,
        mode: GLenumLike | None = None,
        first: int = 0,
        vertices: int | None = None,
        instances: int = 1,
    ) -> None:
        """Render the geometry with a specific program.

        The geometry object will know how many vertices your buffers contains
        so overriding vertices is not needed unless you have a special case
        or have resized the buffers after the geometry instance was created.

        Args:
            program:
                The Program to render with
            mode:
                Override what primitive mode should be used
            first:
                Offset start vertex
            vertices:
                Override the number of vertices to render
            instances:
                Number of instances to render
        """
        program.use()
        vao = self.instance(program)

        mode = self._mode if mode is None else mode

        # If we have a geometry shader we need to sanity check that
        # the primitive mode is supported
        if program.geometry_vertices > 0:
            if program.geometry_input == self._ctx.POINTS:
                mode = program.geometry_input
            if program.geometry_input == self._ctx.LINES:
                if mode not in [
                    self._ctx.LINES,
                    self._ctx.LINE_STRIP,
                    self._ctx.LINE_LOOP,
                    self._ctx.LINES_ADJACENCY,
                ]:
                    raise ValueError(
                        "Geometry shader expects LINES, LINE_STRIP, LINE_LOOP "
                        " or LINES_ADJACENCY as input"
                    )
            if program.geometry_input == self._ctx.LINES_ADJACENCY:
                if mode not in [self._ctx.LINES_ADJACENCY, self._ctx.LINE_STRIP_ADJACENCY]:
                    raise ValueError(
                        "Geometry shader expects LINES_ADJACENCY or LINE_STRIP_ADJACENCY as input"
                    )
            if program.geometry_input == self._ctx.TRIANGLES:
                if mode not in [
                    self._ctx.TRIANGLES,
                    self._ctx.TRIANGLE_STRIP,
                    self._ctx.TRIANGLE_FAN,
                ]:
                    raise ValueError(
                        "Geometry shader expects GL_TRIANGLES, GL_TRIANGLE_STRIP "
                        "or GL_TRIANGLE_FAN as input"
                    )
            if program.geometry_input == self._ctx.TRIANGLES_ADJACENCY:
                if mode not in [self._ctx.TRIANGLES_ADJACENCY, self._ctx.TRIANGLE_STRIP_ADJACENCY]:
                    raise ValueError(
                        "Geometry shader expects GL_TRIANGLES_ADJACENCY or "
                        "GL_TRIANGLE_STRIP_ADJACENCY as input"
                    )

        vao.render(
            mode=mode,
            first=first,
            vertices=vertices or self._num_vertices,
            instances=instances,
        )

    def render_indirect(
        self,
        program: Program,
        buffer: Buffer,
        *,
        mode: GLuintLike | None = None,
        count: int = -1,
        first: int = 0,
        stride: int = 0,
    ) -> None:
        """
        Render the VertexArray to the framebuffer using indirect rendering.

        .. Warning:: This requires OpenGL 4.3

        The following structs are expected for the buffer::

            // Array rendering - no index buffer (16 bytes)
            typedef  struct {
                uint  count;
                uint  instanceCount;
                uint  first;
                uint  baseInstance;
            } DrawArraysIndirectCommand;

            // Index rendering - with index buffer 20 bytes
            typedef  struct {
                GLuint  count;
                GLuint  instanceCount;
                GLuint  firstIndex;
                GLuint  baseVertex;
                GLuint  baseInstance;
            } DrawElementsIndirectCommand;

        The ``stride`` is the byte stride between every rendering command
        in the buffer. By default we assume this is 16 for array rendering
        (no index buffer) and 20 for indexed rendering (with index buffer)

        Args:
            program:
                The program to execute
            buffer:
                The buffer containing one or multiple draw parameters
            mode:
                Primitive type to render. TRIANGLES, LINES etc.
            count:
                The number if indirect draw calls to run.
                If omitted all draw commands in the buffer will be executed.
            first:
                The first indirect draw call to start on
            stride:
                The byte stride of the draw command buffer.
                Keep the default (0) if the buffer is tightly packed.
        """
        program.use()
        vao = self.instance(program)

        mode = self._mode if mode is None else mode
        vao.render_indirect(buffer, mode, count, first, stride)

    def transform(
        self,
        program: Program,
        buffer: Buffer | list[Buffer],
        *,
        first: int = 0,
        vertices: int | None = None,
        instances: int = 1,
        buffer_offset: int = 0,
    ) -> None:
        """
        Render with transform feedback. Instead of rendering to the screen
        or a framebuffer the result will instead end up in the ``buffer`` we supply.

        If a geometry shader is used the output primitive mode is automatically detected.

        Args:
            program:
                The Program to render with
            buffer:
                The buffer(s) we transform into.
                This depends on the programs ``varyings_capture_mode``. We can transform
                into one buffer interleaved or transform each attribute into separate buffers.
            first:
                Offset start vertex
            vertices:
                Number of vertices to render
            instances:
                Number of instances to render
            buffer_offset:
                Byte offset for the buffer
        """
        program.use()
        vao = self.instance(program)
        if program._varyings_capture_mode == "interleaved":
            if not isinstance(buffer, Buffer):
                raise ValueError(
                    (
                        "Buffer must be a single Buffer object "
                        "because the capture mode of the program is: "
                        f"{program.varyings_capture_mode}"
                    )
                )
            vao.transform_interleaved(
                buffer,
                mode=program.geometry_input,
                output_mode=program.geometry_output,
                first=first,
                vertices=vertices or self._num_vertices,
                instances=instances,
                buffer_offset=buffer_offset,
            )
        else:
            if not isinstance(buffer, list):
                raise ValueError(
                    (
                        "buffer must be a list of Buffer object "
                        "because the capture mode of the program is: "
                        f"{program.varyings_capture_mode}"
                    )
                )
            vao.transform_separate(
                buffer,
                mode=program.geometry_input,
                output_mode=program.geometry_output,
                first=first,
                vertices=vertices or self._num_vertices,
                instances=instances,
                buffer_offset=buffer_offset,
            )

    def flush(self) -> None:
        """
        Flush all the internally generated VertexArrays.

        The Geometry instance will store a VertexArray
        for every unique set of input attributes it
        stumbles over when rendering and transform calls
        are issued. This data is usually pretty light weight
        and usually don't need flushing.
        """
        self._vao_cache = {}

    def _generate_vao(self, program: Program) -> VertexArray:
        """
        Create a new VertexArray for the given program.

        Args:
            program: The program to use
        """
        # print(f"Generating vao for key {program.attribute_key}")

        vao = VertexArray(
            self._ctx,
            program,
            self._content,
            index_buffer=self._index_buffer,
            index_element_size=self._index_element_size,
        )
        self._vao_cache[program.attribute_key] = vao
        return vao

    @staticmethod
    def _release(ctx) -> None:
        """Mainly here to count destroyed instances"""
        ctx.stats.decr("geometry")



"""
Camera Example

Artwork from: https://kenney.nl
Tiled available from: https://www.mapeditor.org/

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.camera_platform
"""

import time

import arcade

TILE_SCALING = 0.5
PLAYER_SCALING = 0.5

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720

WINDOW_TITLE = "Camera Example"
SPRITE_PIXEL_SIZE = 128
GRID_PIXEL_SIZE = SPRITE_PIXEL_SIZE * TILE_SCALING

# How many pixels to keep as a minimum margin between the character
# and the edge of the screen.
VIEWPORT_MARGIN_TOP = 60
VIEWPORT_MARGIN_BOTTOM = 60
VIEWPORT_RIGHT_MARGIN = 270
VIEWPORT_LEFT_MARGIN = 270

# Physics
MOVEMENT_SPEED = 5
JUMP_SPEED = 23
GRAVITY = 1.1

# Map Layers
LAYER_NAME_PLATFORMS = "Platforms"
LAYER_NAME_COINS = "Coins"
LAYER_NAME_BOMBS = "Bombs"


class GameView(arcade.View):
    """Main application class."""

    def __init__(self):
        """
        Initializer
        """
        super().__init__()

        # Our TileMap Object
        self.tile_map = None

        # Our Scene Object
        self.scene = None

        # Set up the player
        self.score = 0
        self.player_sprite = None

        self.physics_engine = None
        self.top_of_map = 0
        self.end_of_map = 0
        self.game_over = False
        self.last_time = None
        self.frame_count = 0
        self.fps_message = None

        # Cameras
        self.camera: arcade.Camera2D = None
        self.gui_camera = None

        self.camera_shake = None

        self.shake_offset_1 = 0
        self.shake_offset_2 = 0
        self.shake_vel_1 = 0
        self.shake_vel_2 = 0

        # Text
        self.text_fps = arcade.Text(
            "",
            x=10,
            y=40,
            color=arcade.color.BLACK,
            font_size=14,
        )
        self.text_score = arcade.Text(
            f"Score: {self.score}",
            x=10,
            y=20,
            color=arcade.color.BLACK,
            font_size=14,
        )

    def setup(self):
        """Set up the game and initialize the variables."""

        # Map name
        map_name = ":resources:tiled_maps/level_1.json"

        # Layer Specific Options for the Tilemap
        layer_options = {
            LAYER_NAME_PLATFORMS: {
                "use_spatial_hash": True,
            },
            LAYER_NAME_COINS: {
                "use_spatial_hash": True,
            },
            LAYER_NAME_BOMBS: {
                "use_spatial_hash": True,
            },
        }

        # Load in TileMap
        self.tile_map = arcade.load_tilemap(map_name, TILE_SCALING, layer_options)

        # Initiate New Scene with our TileMap, this will automatically add all layers
        # from the map as SpriteLists in the scene in the proper order.
        self.scene = arcade.Scene.from_tilemap(self.tile_map)

        # Set up the player
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=PLAYER_SCALING,
        )

        # Starting position of the player
        self.player_sprite.center_x = 196
        self.player_sprite.center_y = 128
        self.scene.add_sprite("Player", self.player_sprite)

        self.camera = arcade.Camera2D()

        self.camera_shake = arcade.camera.grips.ScreenShake2D(self.camera.view_data,
                                                              max_amplitude=12.5,
                                                              acceleration_duration=0.05,
                                                              falloff_time=0.20,
                                                              shake_frequency=15.0)

        # Center camera on user
        self.pan_camera_to_user()

        # Calculate the right edge of the my_map in pixels
        self.top_of_map = self.tile_map.height * GRID_PIXEL_SIZE
        self.end_of_map = self.tile_map.width * GRID_PIXEL_SIZE

        # --- Other stuff
        # Set the background color
        if self.tile_map.background_color:
            self.background_color = self.tile_map.background_color

        # Keep player from running through the wall_list layer
        self.physics_engine = arcade.PhysicsEnginePlatformer(
            self.player_sprite,
            self.scene.get_sprite_list(LAYER_NAME_PLATFORMS),
            gravity_constant=GRAVITY,
        )

        self.game_over = False

    def on_resize(self, width, height):
        """Resize window"""
        super().on_resize(width, height)
        self.camera.match_window()

    def on_draw(self):
        """Render the screen."""
        self.clear()

        self.camera_shake.update_camera()
        with self.camera.activate():
            # Draw our Scene
            self.scene.draw()
        # Readjust the camera so the screen shake doesn't affect
        # the camera following algorithm.
        self.camera_shake.readjust_camera()

        with self.window.default_camera.activate():
            # Update fps text periodically
            if self.last_time and self.frame_count % 60 == 0:
                fps = 1.0 / (time.time() - self.last_time) * 60
                self.text_fps.text = f"FPS: {fps:5.2f}"

            self.text_fps.draw()

            if self.frame_count % 60 == 0:
                self.last_time = time.time()

            # Draw Score
            self.text_score.draw()

            # Draw game over
            if self.game_over:
                arcade.draw_text("Game Over", self.width/2, self.height/2, arcade.color.BLACK,
                                 30)

        self.frame_count += 1

    def on_key_press(self, key, modifiers):
        """
        Called whenever a key is pressed
        """
        if key == arcade.key.UP:
            if self.physics_engine.can_jump():
                self.player_sprite.change_y = JUMP_SPEED
        elif key == arcade.key.LEFT:
            self.player_sprite.change_x = -MOVEMENT_SPEED
        elif key == arcade.key.RIGHT:
            self.player_sprite.change_x = MOVEMENT_SPEED

    def on_key_release(self, key, modifiers):
        """
        Called when the user presses a mouse button.
        """
        if key == arcade.key.LEFT or key == arcade.key.RIGHT:
            self.player_sprite.change_x = 0

    def pan_camera_to_user(self, panning_fraction: float = 1.0):
        """
        Manage Scrolling

        Args:
            panning_fraction:
                Number from 0 to 1. Higher the number, faster we
                pan the camera to the user.
        """

        # This spot would center on the user
        screen_center_x, screen_center_y = self.player_sprite.position
        if screen_center_x < self.camera.viewport_width/2:
            screen_center_x = self.camera.viewport_width/2
        if screen_center_y < self.camera.viewport_height/2:
            screen_center_y = self.camera.viewport_height/2
        user_centered = screen_center_x, screen_center_y

        self.camera.position = arcade.math.lerp_2d(
            self.camera.position,
            user_centered,
            panning_fraction,
        )

    def on_update(self, delta_time):
        """Movement and game logic"""

        if self.player_sprite.right >= self.end_of_map:
            self.game_over = True

        # Call update on all sprites
        if not self.game_over:
            self.physics_engine.update()
            self.camera_shake.update(delta_time)

        coins_hit = arcade.check_for_collision_with_list(
            self.player_sprite, self.scene.get_sprite_list("Coins")
        )
        for coin in coins_hit:
            coin.remove_from_sprite_lists()
            self.score += 1

        # Bomb hits
        bombs_hit = arcade.check_for_collision_with_list(
            self.player_sprite, self.scene.get_sprite_list("Bombs")
        )
        for bomb in bombs_hit:
            bomb.remove_from_sprite_lists()
            self.camera_shake.start()

        # Pan to the user
        self.pan_camera_to_user(panning_fraction=0.12)

        # Update score text
        self.text_score.text = f"Score: {self.score}"


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""Packets of data and base types supporting cameras.

These are placed in their own module to simplify imports due to their
wide usage throughout Arcade's camera code.
"""

from collections.abc import Generator
from contextlib import contextmanager
from typing import Final, Protocol

from pyglet.math import Vec2, Vec3
from typing_extensions import Self

from arcade.types import LRBT, AsFloat, Point, Point3, Rect

__all__ = [
    "CameraData",
    "DEFAULT_FAR",
    "DEFAULT_NEAR_ORTHO",
    "OrthographicProjectionData",
    "PerspectiveProjectionData",
    "Projection",
    "Projector",
    "ZeroProjectionDimension",
    "constrain_camera_data",
    "duplicate_camera_data",
]

DEFAULT_NEAR_ORTHO: Final[float] = -100.0
"""The default backward-facing depth cutoff for orthographic rendering.

Unless an orthographic camera is provided a different value, this will be
used as the near cutoff for its point of view.

The :py:class:`~arcade.camera.perspective.PerspectiveProjector` uses
``0.01`` as its default near value to avoid division by zero.
"""

DEFAULT_FAR: Final[float] = 100.0
"""The default forward-facing depth cutoff for all Arcade cameras.

Unless a camera is provided a different value, anything further away than this
value will not be drawn.
"""


class ZeroProjectionDimension(ValueError):
    """A projection's dimensions were zero along at least one axis.

    This usually happens because code tried to set one of the following:

    * ``left`` equal to ``right``
    * ``bottom`` equal to ``top``

    You can handle this error as a :py:class:`ValueError`.
    """

    ...


class CameraData:
    """Stores position, orientation, and zoom for a camera.

    This is like where a camera is placed in 3D space.

    Args:
        position:
            The camera's location in 3D space.
        up:
            The direction which is considered "up" for the camera.
        forward:
            The direction the camera is facing.
        zoom:
            How much the camera is zoomed in or out.
    """

    __slots__ = ("position", "up", "forward", "zoom")

    def __init__(
        self,
        position: Point3 = (0.0, 0.0, 0.0),
        up: Point3 = (0.0, 1.0, 0.0),
        forward: Point3 = (0.0, 0.0, -1.0),
        zoom: float = 1.0,
    ):
        self.position: tuple[float, float, float] = position
        """A 3D vector which describes where the camera is located."""

        self.up: tuple[float, float, float] = up
        """A 3D vector which describes which direction is up (+y)."""

        self.forward: tuple[float, float, float] = forward
        """
        A scalar which describes which direction the camera is pointing.

        While this affects the projection matrix, it also allows camera
        controllers to access zoom functionality without interacting with
        projection data.
        """

        self.zoom: float = zoom
        """A scalar which describes how much the camera is zoomed in or out."""

    def __str__(self):
        return f"CameraData<{self.position=}, {self.up=}, {self.forward=}, {self.zoom=}>"

    def __repr__(self):
        return self.__str__()


def duplicate_camera_data(origin: CameraData):
    """
    Clone camera data

    Args:
        origin: The camera data to clone
    """
    return CameraData(origin.position, origin.up, origin.forward, float(origin.zoom))


def constrain_camera_data(data: CameraData, forward_priority: bool = False):
    """
    Ensure that the camera data forward and up vectors are length one,
    and are perpendicular

    Args:
        data: the camera data to constrain
        forward_priority: whether up or forward gets constrained
    """
    forward_vec = Vec3(*data.forward).normalize()
    up_vec = Vec3(*data.up).normalize()
    right_vec = forward_vec.cross(up_vec).normalize()
    if forward_priority:
        up_vec = right_vec.cross(forward_vec)
    else:
        forward_vec = up_vec.cross(right_vec)

    data.forward = (forward_vec.x, forward_vec.y, forward_vec.z)
    data.up = (up_vec.x, up_vec.y, up_vec.z)


class OrthographicProjectionData:
    """Describes an Orthographic projection.

    This is by default a Left-handed system. with the X axis going from left to
    right, The Y axis going from bottom to top, and the Z axis going from towards
    the screen to away from the screen. This can be made right-handed by making
    the near value greater than the far value.

    Args:
        left: Left limit of the projection
        right: Right limit of the projection
        bottom: Bottom limit of the projection
        top: Top limit of the projection
        near: Near plane
        far: Far plane
    """

    __slots__ = ("rect", "near", "far")

    def __init__(
        self, left: float, right: float, bottom: float, top: float, near: float, far: float
    ):
        self.rect: Rect = LRBT(left, right, bottom, top)
        """Rectangle defining the projection area."""

        self.near: float = near
        """
        The 'closest' visible position along the forward direction.

        It will get mapped to z = -1.0. Anything closer than this value
        is not visible.
        """

        self.far: float = far
        """
        The 'farthest' visible position along the forward direction.

        It will get mapped to z = 1.0. Anything father than this value
        is not visible.
        """

    @property
    def left(self) -> float:
        """
        The left-side cutoff value, which gets mapped to x = -1.0.

        Anything to the left of this value is not visible.
        """
        return self.rect.left

    @left.setter
    def left(self, new_left: AsFloat):
        r = self.rect
        dl = new_left - r.left
        self.rect = Rect(
            new_left, r.right, r.bottom, r.top, r.width + dl, r.height, r.x + dl / 2.0, r.y
        )

    @property
    def right(self) -> float:
        """
        The right-side cutoff value, which gets mapped to x = 1.0.

        Anything to the left of this value is not visible.
        """
        return self.rect.right

    @right.setter
    def right(self, new_right: AsFloat):
        r = self.rect
        dr = new_right - r.right
        self.rect = Rect(
            r.left, new_right, r.bottom, r.top, r.width + dr, r.height, r.x + dr / 2.0, r.y
        )

    @property
    def bottom(self) -> float:
        """
        The bottom-side cutoff value, which gets mapped to -y = 1.0.

        Anything to the left of this value is not visible.
        """
        return self.rect.bottom

    @bottom.setter
    def bottom(self, new_bottom: AsFloat):
        r = self.rect
        db = new_bottom - r.bottom
        self.rect = Rect(
            r.left, r.right, new_bottom, r.top, r.width, r.height + db, r.x, r.y + db / 2.0
        )

    @property
    def top(self) -> float:
        """
        The top-side cutoff value, which gets mapped to y = 1.0.

        Anything to the left of this value is not visible.
        """
        return self.rect.top

    @top.setter
    def top(self, new_top: AsFloat):
        r = self.rect
        dt = new_top - r.top
        self.rect = Rect(
            r.left, r.right, r.bottom, new_top, r.width, r.height + dt, r.x, r.y + dt / 2.0
        )

    @property
    def lrbt(self) -> tuple[float, float, float, float]:
        """The left, right, bottom, and top values of the projection."""
        return self.rect.lrbt

    @lrbt.setter
    def lrbt(self, new_lrbt: tuple[float, float, float, float]):
        self.rect = LRBT(*new_lrbt)

    def __str__(self):
        return f"OrthographicProjection<LRBT={self.rect.lrbt}, {self.near=}, {self.far=}>"

    def __repr__(self):
        return self.__str__()


def orthographic_from_rect(rect: Rect, near: float, far: float) -> OrthographicProjectionData:
    """
    Create an orthographic projection from a rectangle.

    Args:
        rect: The rectangle to create the projection from.
        near: The near plane of the projection.
        far: The far plane of the projection.
    """
    return OrthographicProjectionData(rect.left, rect.right, rect.bottom, rect.top, near, far)


class PerspectiveProjectionData:
    """
    Data for perspective projection.

    Args:
        aspect: The aspect ratio of the screen (width over height).
        fov: The field of view in degrees.
        near: The 'closest' visible position along the forward direction.
        far: The 'farthest' visible position along the forward
    """

    __slots__ = ("aspect", "fov", "near", "far")

    def __init__(self, aspect: float, fov: float, near: float, far: float):
        self.aspect: float = aspect
        """The aspect ratio of the screen (width over height)."""

        self.fov: float = fov
        """
        The field of view in degrees.

        Together with the aspect ratio, it defines the size of the
        perspective projection for any given depth.
        """

        self.near: float = near
        """
        The 'closest' visible position along the forward direction.

        It will get mapped to z = -1.0. Anything closer than this value
        is not visible.
        """

        self.far: float = far
        """"
        The 'farthest' visible position along the forward direction.

        It will get mapped to z = 1.0. Anything father than this value
        is not visible.
        """

    def __str__(self) -> str:
        return f"PerspectiveProjection<{self.aspect=}, {self.fov=}, {self.near=}, {self.far=}>"

    def __repr__(self) -> str:
        return self.__str__()


class Projection(Protocol):
    """Matches the data universal in Arcade's projection data objects.

    There are multiple types of projections used in games, but all the
    common ones share key features. This :py:class:`~typing.Protocol`:

    #. Defines those shared elements
    #. Annotates these in code for both humans and automated type
       checkers

    The specific implementations which match it are used inside of
    implementations of Arcade's :py:class:`.Projector` behavior. All
    of these projectors rely on a ``viewport`` as well as ``near`` and
    ``far`` values.

    The ``viewport`` is measured in screen pixels. By default, the
    conventions for this are the same as the rest of Arcade and
    OpenGL:

    * X is measured rightward from left of the screen
    * Y is measured up from the bottom of the screen

    Although the ``near`` and ``far`` values are describe the cutoffs
    for what the camera sees in world space, the exact meaning differs
    between projection type.

    .. list-table::
       :header-rows: 1

       * - Common Projection Type
         - Meaning of ``near`` & ``far``

       * - Simple Orthographic
         - The Z position in world space

       * - Perspective & Isometric
         - Where the rear and front clipping planes sit along a
           camera's :py:attr:`.CameraData.forward` vector.

    """

    near: float
    far: float


class Projector(Protocol):
    """Projects from world coordinates to viewport pixel coordinates.

    Projectors also support converting in the opposite direction from
    screen pixel coordinates to world space coordinates.

    The two key spatial methods which do this are:

    .. list-table::
       :header-rows: 1

       * - Method
         - Action

       * - :py:meth:`.project`
         - Turn world coordinates into pixel coordinates relative
           to the origin (bottom left by default).

       * - :py:meth:`.unproject`
         - Convert screen pixel coordinates into world space.

    .. note: Every :py:class:`.Camera` is also a kind of projector.

    The other required methods are for helping manage which camera is
    currently used to draw.

    """

    def use(self) -> None:
        """Set the GL context to use this projector and its settings.

        .. warning:: You may be looking for:py:meth:`.activate`!

                     This method only sets rendering state for a given
                     projector. Since it doesn't restore any afterward,
                     it's easy to misuse in ways which can cause bugs
                     or temporarily break a game's rendering until
                     relaunch. For reliable, automatic clean-up see
                     the :py:meth:`.activate` method instead.

        If you are implementing your own custom projector, this method
        should only:

        #. Set the Arcade :py:class:`~arcade.Window`'s
           :py:attr:`~arcade.Window.current_camera` to this object
        #. Calculate any required view and projection matrices
        #. Set any resulting values on the current
           :py:class:`~arcade.ArcadeContext`, including the:

           * :py:attr:`~arcade.ArcadeContext.viewport`
           * :py:attr:`~arcade.ArcadeContext.view_matrix`
           * :py:attr:`~arcade.ArcadeContext.projection_matrix`

        This method should **never** handle cleanup. That is the
        responsibility of :py:attr:`.activate`.
        """
        ...

    @contextmanager
    def activate(self) -> Generator[Self, None, None]:
        """
        Activate this projector for rendering.

        This is a context manager and should be used with a ``with`` statement::

            with projector.activate():
                # Render with this projector
        """
        ...

    def project(self, world_coordinate: Point) -> Vec2:
        """
        Take a Vec2 or Vec3 of coordinates and return the related screen coordinate
        """
        ...

    def unproject(self, screen_coordinate: Point) -> Vec3:
        """
        Take in a pixel coordinate and return the associated world coordinate

        Essentially reverses the effects of the projector.

        Args:
            screen_coordinate: A 2D position in pixels should generally be inside
                the range of the active viewport.
        Returns:
            A 3D vector in world space.
        """
        ...



"""This example shows how to use arcade.gui with a camera.
It is a simple game where the player can move around and collect coins.
The player can upgrade their speed and the spawn rate of the coins.
The game has a timer and ends after 60 seconds.
The game is controlled with the arrow keys or WASD.

At the beginning of the game, the UI camera is used, to apply some animations.

If Arcade and Python are properly installed, you can run this example with:
python -m arcade.examples.gui.4_with_camera
"""

import math
import random

import arcade
from arcade.gui import UIAnchorLayout, UIBoxLayout, UIFlatButton, UILabel, UIOnClickEvent, UIView

COIN_PNG = ":resources:images/items/coinGold.png"
ADV_PNG = ":resources:/images/animated_characters/female_adventurer/femaleAdventurer_idle.png"


class MyCoinGame(UIView):
    """Main view of the game. This class is a subclass of UIView, which provides
    basic GUI setup. We add UIManager to the view under `self.ui`.

    The example showcases how to:
    - use UIView to set up a basic GUI
    - add a button to the view and connect it to a function
    - use camera to move the view

    """

    def __init__(self):
        super().__init__()
        self.bg_color = arcade.uicolor.DARK_BLUE_MIDNIGHT_BLUE

        # basic camera setup
        self.keys = set()
        self.in_game_camera = arcade.Camera2D()
        self.in_game_camera.bottom_left = 100, 100

        # in-game counter
        self._total_time = 0
        self._game_duration = 60
        self._game_over = False
        self._last_coin_spawn = 0
        self._coin_spawn_delay = 3
        self._coins_collected = 0

        # upgradable player values
        self._player_speed = 5

        # setup in-game objects
        self.sprites = arcade.SpriteList()

        self.game_area = arcade.SpriteSolidColor(
            width=1300,
            height=730,
            color=arcade.color.GRAY_ASPARAGUS,
            center_x=1280 / 2,
            center_y=720 / 2,
        )

        self.sprites.append(self.game_area)

        self.player = arcade.Sprite(
            ADV_PNG,
            scale=0.5,
            center_x=1280 / 2,
            center_y=720 / 2,
        )
        self.sprites.append(self.player)

        self.coins = arcade.SpriteList()
        for i in range(12):
            # place coins in a circle around the player, radius =100
            coin = arcade.Sprite(
                COIN_PNG,
                scale=0.5,
                center_x=1280 / 2 + 200 * math.cos(math.radians(i * 40)),
                center_y=720 / 2 + 200 * math.sin(math.radians(i * 40)),
            )
            self.coins.append(coin)

        # UI setup, we use UIView, which automatically adds UIManager as self.ui
        anchor = self.ui.add(UIAnchorLayout())

        shop_buttons = anchor.add(
            UIBoxLayout(vertical=False, space_between=10),
            anchor_x="center",
            anchor_y="bottom",
            align_y=10,
        )

        # speed upgrade button
        speed_upgrade = UIFlatButton(text="Upgrade Speed (5C)", width=200, height=40)
        shop_buttons.add(speed_upgrade)

        @speed_upgrade.event("on_click")
        def upgrade_speed(event: UIOnClickEvent):
            cost = self._player_speed
            if self._coins_collected >= cost:
                self._coins_collected -= cost
                self._player_speed += 1
                speed_upgrade.text = f"Update Speed ({self._player_speed}C)"
                print("Speed upgraded")

        # update spawn rate button
        spawn_rate_upgrade = UIFlatButton(text="Upgrade spawn rate: 10C", width=300, height=40)
        shop_buttons.add(spawn_rate_upgrade)

        @spawn_rate_upgrade.event("on_click")
        def upgrade_spawn_rate(event: UIOnClickEvent):
            cost = 10
            if self._coins_collected >= cost:
                self._coins_collected -= cost
                self._coin_spawn_delay -= 0.5
                print("Spawn rate upgraded")

        # position top center, with a 40px offset
        self.out_of_game_area = anchor.add(
            UILabel(text="Out of game area", font_size=32),
            anchor_x="center",
            anchor_y="top",
            align_y=-40,
        )
        self.out_of_game_area.visible = False

        self.coin_counter = anchor.add(
            UILabel(text="Collected coins 0", size_hint=(0, 0)),
            anchor_x="left",
            anchor_y="top",
            align_y=-10,
            align_x=10,
        )
        self.coin_counter.with_background(
            color=arcade.color.TRANSPARENT_BLACK
            # giving a background will make the label way more performant,
            # because it will not re-render the whole UI after a text change
        )

        # Game timer
        self.timer = anchor.add(
            UILabel(
                text="Time 30.0",
                font_size=15,
                size_hint=(0, 0),  # take the whole width to prevent linebreaks
            ),
            anchor_x="center",
            anchor_y="top",
            align_y=-10,
            align_x=-10,
        )
        self.timer.with_background(color=arcade.color.TRANSPARENT_BLACK)

        self.cam_pos = self.ui.camera.position

    def on_draw_before_ui(self):
        self.in_game_camera.use()  # use the in-game camera to draw in-game objects
        self.sprites.draw()
        self.coins.draw()

    def on_update(self, delta_time: float) -> bool | None:
        if self._total_time > self._game_duration:
            # ad new UI label to show the end of the game
            game_over_text = self.ui.add(
                UILabel(
                    text="End of game!\n"
                    f"You achieved {self._coins_collected} coins!\n"
                    "Press ESC to exit.\n"
                    "Use ENTER to restart.",
                    font_size=32,
                    bold=True,
                    multiline=True,
                    align="center",
                    text_color=arcade.color.WHITE,
                    size_hint=(0, 0),
                ),
            )
            game_over_text.with_padding(all=10)
            game_over_text.with_background(color=arcade.types.Color(50, 50, 50, 120))
            game_over_text.center_on_screen()

            return True

        self._total_time += delta_time
        self._last_coin_spawn += delta_time

        # update the timer
        self.timer.text = f"Time {self._game_duration - self._total_time:.1f}"

        # spawn new coins
        if self._last_coin_spawn > self._coin_spawn_delay:
            coin = arcade.Sprite(
                COIN_PNG,
                scale=0.5,
                center_x=random.randint(0, 1280),
                center_y=random.randint(0, 720),
            )
            self.coins.append(coin)
            self._last_coin_spawn -= self._coin_spawn_delay

        # move the player sprite
        if {arcade.key.LEFT, arcade.key.A} & self.keys:
            self.player.left -= self._player_speed
        if {arcade.key.RIGHT, arcade.key.D} & self.keys:
            self.player.left += self._player_speed
        if {arcade.key.UP, arcade.key.W} & self.keys:
            self.player.top += self._player_speed
        if {arcade.key.DOWN, arcade.key.S} & self.keys:
            self.player.top -= self._player_speed

        # move the camera with the player
        self.in_game_camera.position = self.player.position

        # collect coins
        collisions = self.player.collides_with_list(self.coins)
        for coin in collisions:
            coin.remove_from_sprite_lists()
            self._coins_collected += 1
            print("Coin collected")

        # update the coin counter
        self.coin_counter.text = f"Collected coins {self._coins_collected}"

        # inform player if they are out of the game area
        if not self.player.collides_with_sprite(self.game_area):
            self.out_of_game_area.visible = True
        else:
            self.out_of_game_area.visible = False

        # slide in the UI from bottom, until total time reaches 2 seconds
        progress = min(1.0, self._total_time / 2)

        # Because we allow for camera rotation we have work in the center
        # and not the edge because it behaves oddly otherwise
        cam_pos_x = self.window.center_x
        cam_pos_y = 50 * (1 - progress) + self.window.center_y
        self.ui.camera.position = (cam_pos_x, cam_pos_y)

        return False

    def on_key_press(self, symbol: int, modifiers: int) -> bool | None:
        self.keys.add(symbol)

        if symbol == arcade.key.ESCAPE:
            arcade.close_window()
        if symbol == arcade.key.ENTER:
            self.window.show_view(MyCoinGame())

        return False

    def on_key_release(self, symbol: int, modifiers: int) -> bool | None:
        if symbol in self.keys:
            self.keys.remove(symbol)
        return False


def main():
    window = arcade.Window(1280, 720, "GUI Example: Coin Game (Camera)", resizable=False)
    window.show_view(MyCoinGame())
    window.run()


if __name__ == "__main__":
    main()



"""
Move Sprite with Controller

An example of one way to use controller and controller input to move a
sprite.

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the
command line with:
python -m arcade.examples.sprite_move_controller
"""
import arcade

SPRITE_SCALING = 0.5

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Move Sprite with Controller Example"

MOVEMENT_SPEED = 5
DEAD_ZONE = 0.05


class Player(arcade.Sprite):
    """ Player sprite """

    def __init__(self, filename, scale):
        super().__init__(filename, scale=scale)

        self.controller = None

        controllers = arcade.get_controllers()

        # If we have any...
        if controllers:
            # Grab the first one in  the list
            self.controller = controllers[0]

            # Open it for input
            self.controller.open()

            # Push this object as a handler for controller events.
            # Required for the controller events to be called.
            self.controller.push_handlers(self)

    def update(self, delta_time: float = 1 / 60):
        """ Move the player """

        # If there is a controller, grab the speed.
        if self.controller:

            # x-axis
            movement_x = self.controller.leftx
            if abs(movement_x) < DEAD_ZONE:
                self.change_x = 0
            else:
                self.change_x = movement_x * MOVEMENT_SPEED

            # y-axis
            movement_y = self.controller.lefty
            if abs(movement_y) < DEAD_ZONE:
                self.change_y = 0
            else:
                self.change_y = movement_y * MOVEMENT_SPEED

        # Move the player
        self.center_x += self.change_x
        self.center_y += self.change_y

        # Keep from moving off-screen
        if self.left < 0:
            self.left = 0
        elif self.right > WINDOW_WIDTH - 1:
            self.right = WINDOW_WIDTH - 1

        if self.bottom < 0:
            self.bottom = 0
        elif self.top > WINDOW_HEIGHT - 1:
            self.top = WINDOW_HEIGHT - 1

    def on_button_press(self, controller, button_name):
        """ Handle button-down event for the controller """
        print(f"Button {button_name} down")

    def on_button_release(self, controller, button_name):
        """ Handle button-up event for the controller """
        print(f"Button {button_name} up")

    def on_stick_motion(self, controller, stick_name, x, y):
        """ Handle hat events """
        print(f"Movement on stick {stick_name}: ({x}, {y})")


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        """
        Initializer
        """
        # Call the parent class initializer
        super().__init__()

        # Variables that will hold sprite lists
        self.all_sprites_list = None

        # Set up the player info
        self.player_sprite = None

        # Set the background color
        self.background_color = arcade.color.AMAZON

        self.error_text = arcade.Text(
            "There are no controllers, plug in a controller and run again.",
            10,
            10,
            arcade.color.WHITE,
            22,
            width=WINDOW_WIDTH,
            align="center",
        )

    def setup(self):
        """ Set up the game and initialize the variables. """

        # Sprite lists
        self.all_sprites_list = arcade.SpriteList()

        # Set up the player
        self.player_sprite = Player(
            ":resources:images/animated_characters/female_person/"
            "femalePerson_idle.png",
            SPRITE_SCALING,
        )
        self.player_sprite.position = self.width / 2, self.height / 2
        self.all_sprites_list.append(self.player_sprite)

    def on_draw(self):
        """
        Render the screen.
        """

        # This command has to happen before we start drawing
        self.clear()

        # Draw all the sprites.
        self.all_sprites_list.draw()

        # Print an error if there is no controller
        if not self.player_sprite.controller:
            self.error_text.draw()

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Call update on all sprites (The sprites don't do much in this
        # example though.)
        self.all_sprites_list.update(delta_time)

    def on_key_press(self, key, modifiers):
        """Called whenever a key is pressed. """

        if key == arcade.key.UP:
            self.player_sprite.change_y = MOVEMENT_SPEED
        elif key == arcade.key.DOWN:
            self.player_sprite.change_y = -MOVEMENT_SPEED
        elif key == arcade.key.LEFT:
            self.player_sprite.change_x = -MOVEMENT_SPEED
        elif key == arcade.key.RIGHT:
            self.player_sprite.change_x = MOVEMENT_SPEED

    def on_key_release(self, key, modifiers):
        """Called when the user releases a key. """

        if key == arcade.key.UP or key == arcade.key.DOWN:
            self.player_sprite.change_y = 0
        elif key == arcade.key.LEFT or key == arcade.key.RIGHT:
            self.player_sprite.change_x = 0


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



from PIL.Image import Image

from arcade.types import Point2List

from .base import HitBoxAlgorithm


class BoundingHitBoxAlgorithm(HitBoxAlgorithm):
    """
    A simple hit box algorithm that returns a hit box around the entire image.
    """

    cache = False

    def calculate(self, image: Image, **kwargs) -> Point2List:
        """
        Given an RGBA image, this returns points that make up a hit box around it
        without any attempt to trim out transparent pixels.

        Args:
            image: Image get hit box from.
        """
        return self.create_bounding_box(image)



from __future__ import annotations

from collections.abc import Generator
from contextlib import contextmanager
from typing import TYPE_CHECKING

from pyglet.math import Mat4, Vec2, Vec3
from typing_extensions import Self

from arcade.camera.data_types import (
    DEFAULT_FAR,
    DEFAULT_NEAR_ORTHO,
    CameraData,
    OrthographicProjectionData,
    Projector,
)
from arcade.camera.projection_functions import (
    generate_orthographic_matrix,
    generate_view_matrix,
    project_orthographic,
    unproject_orthographic,
)
from arcade.types import LBWH, Point, Rect
from arcade.window_commands import get_window

if TYPE_CHECKING:
    from arcade import Window


__all__ = ["OrthographicProjector"]


class OrthographicProjector(Projector):
    """
    The simplest from of an orthographic camera.
    Using ViewData and OrthographicProjectionData PoDs (Pack of Data)
    it generates the correct projection and view matrices. It also
    provides methods and a context manager for using the matrices in
    glsl shaders.

    This class provides no methods for manipulating the PoDs.

    The current implementation will recreate the view and
    projection matrices every time the camera is used.
    If used every frame or multiple times per frame this may
    be inefficient. If you suspect this is causing slowdowns
    profile before optimizing with a dirty value check.

    Initialize a Projector which produces an orthographic projection matrix using
    a CameraData and PerspectiveProjectionData PoDs.

    Args:
        window:
            The window to bind the camera to. Defaults to the currently active camera.
        view:
            The CameraData PoD. contains the viewport, position, up, forward, and zoom.
        projection:
            The OrthographicProjectionData PoD.
            contains the left, right, bottom top, near, and far planes.
    """

    def __init__(
        self,
        *,
        window: Window | None = None,
        view: CameraData | None = None,
        projection: OrthographicProjectionData | None = None,
        viewport: Rect | None = None,
        scissor: Rect | None = None,
    ):
        self._window: Window = window or get_window()

        self.viewport: Rect = viewport or LBWH(0, 0, self._window.width, self._window.height)
        self.scissor: Rect | None = scissor

        self._view = view or CameraData(  # Viewport
            (self._window.width / 2, self._window.height / 2, 0),  # Position
            (0.0, 1.0, 0.0),  # Up
            (0.0, 0.0, -1.0),  # Forward
            1.0,  # Zoom
        )

        self._projection = projection or OrthographicProjectionData(
            -0.5 * self._window.width,
            0.5 * self._window.width,  # Left, Right
            -0.5 * self._window.height,
            0.5 * self._window.height,  # Bottom, Top
            DEFAULT_NEAR_ORTHO,
            DEFAULT_FAR,  # Near, Far
        )

    @property
    def view(self) -> CameraData:
        """
        The CameraData. Is a read only property.
        """
        return self._view

    @property
    def projection(self) -> OrthographicProjectionData:
        """
        The OrthographicProjectionData. Is a read only property.
        """
        return self._projection

    def generate_projection_matrix(self) -> Mat4:
        """
        alias of arcade.camera.get_orthographic_matrix method
        """
        return generate_orthographic_matrix(self._projection, self._view.zoom)

    def generate_view_matrix(self) -> Mat4:
        """
        alias of arcade.camera.get_view_matrix method
        """
        return generate_view_matrix(self._view)

    def use(self) -> None:
        """
        Sets the active camera to this object.
        Then generates the view and projection matrices.
        Finally, the gl context viewport is set, as well as the projection and view matrices.
        """

        self._window.current_camera = self

        _projection = generate_orthographic_matrix(self._projection, self._view.zoom)
        _view = generate_view_matrix(self._view)

        self._window.ctx.viewport = self.viewport.lbwh_int
        self._window.ctx.scissor = None if not self.scissor else self.scissor.lbwh_int
        self._window.projection = _projection
        self._window.view = _view

    @contextmanager
    def activate(self) -> Generator[Self, None, None]:
        """Set this camera as the current one, then undo it after.

        This method is a :external:ref:`context manager <context-managers>`
        you can use inside ``with`` blocks. Using it this way guarantees
        that the old camera and its settings will be restored, even if an
        exception occurs:

        .. code-block:: python

           # Despite an Exception, the previous camera and its settings
           # will be restored at the end of the with block below:
           with projector_instance.activate():
                sprite_list.draw()
                _ = 1 / 0  # Guaranteed ZeroDivisionError

        """
        previous_projector = self._window.current_camera
        try:
            self.use()
            yield self
        finally:
            previous_projector.use()

    def project(self, world_coordinate: Point) -> Vec2:
        """
        Take a Vec2 or Vec3 of coordinates and return the related screen coordinate
        """
        _projection = generate_orthographic_matrix(self._projection, self._view.zoom)
        _view = generate_view_matrix(self._view)

        return project_orthographic(
            world_coordinate,
            self.viewport.lbwh_int,
            _view,
            _projection,
        )

    def unproject(self, screen_coordinate: Point) -> Vec3:
        """
        Take in a pixel coordinate from within
        the range of the window size and returns
        the world space coordinates.

        Essentially reverses the effects of the projector.

        Args:
            screen_coordinate: A 2D position in pixels from the bottom left of the screen.
                               This should ALWAYS be in the range of 0.0 - screen size.
        Returns:
            A 3D vector in world space.
        """
        _projection = generate_orthographic_matrix(self._projection, self._view.zoom)
        _view = generate_view_matrix(self._view)
        return unproject_orthographic(screen_coordinate, self.viewport.lbwh_int, _view, _projection)



from PIL import Image

import arcade.gl as gl
from arcade import get_window
from arcade.resources import resolve

from arcade.future.background.background_texture import BackgroundTexture
from arcade.future.background.background import Background
from arcade.future.background.groups import BackgroundGroup, ParallaxGroup

__all__ = [
    "Background",
    "BackgroundGroup",
    "BackgroundTexture",
    "ParallaxGroup",
    "texture_from_file",
    "background_from_file",
]


def texture_from_file(
    tex_src: str,
    offset: tuple[float, float] = (0.0, 0.0),
    scale: float = 1.0,
    angle: float = 0.0,
    filters=(gl.NEAREST, gl.NEAREST),
) -> BackgroundTexture:
    _context = get_window().ctx

    with Image.open(resolve(tex_src)).convert("RGBA") as img:
        texture = _context.texture(
            img.size,
            data=img.transpose(Image.Transpose.FLIP_TOP_BOTTOM).tobytes(),
            filter=filters,
        )

    return BackgroundTexture(texture, offset, scale, angle)


def background_from_file(
    tex_src: str,
    pos: tuple[float, float] = (0.0, 0.0),
    size: tuple[int, int] | None = None,
    offset: tuple[float, float] = (0.0, 0.0),
    scale: float = 1.0,
    angle: float = 0.0,
    *,
    filters=(gl.NEAREST, gl.NEAREST),
    color: tuple[int, int, int] | None = None,
    color_norm: tuple[float, float, float] | None = None,
    shader: gl.Program | None = None,
    geometry: gl.Geometry | None = None,
) -> Background:
    texture = BackgroundTexture.from_file(tex_src, offset, scale, angle, filters)
    if size is None:
        size = texture.texture.size

    if color_norm:
        _color = color_norm
    elif color:
        _color = color[0] / 255, color[1] / 255, color[2] / 255
    else:
        _color = (1.0, 1.0, 1.0)

    return Background(texture, pos, size, _color, shader, geometry)



from PIL import Image
from pyglet.math import Mat3

import arcade.gl as gl
from arcade import ArcadeContext
from arcade.resources import resolve
from arcade.window_commands import get_window


class BackgroundTexture:
    """
    BackgroundTextures are PODs (packet of data) types. They have very little
    functionality by themselves, but are used by Backgrounds. They hold an
    ``arcade.gl.Texture``and 3 `pyglet.math.Mat3``.

    The Mat3s define the scaling, rotation, and translation of the pixel data in the texture.
    see background_fs.glsl in resources/shaders for an implementation of this.

    Args:
        texture: The texture to use as the background.
        offset: The offset of the texture in pixels.
        scale: The scale of the texture.
        angle: The angle of the texture in radians.
    """

    def __init__(
        self,
        texture: gl.Texture2D,
        offset: tuple[float, float] = (0.0, 0.0),
        scale: float = 1.0,
        angle: float = 0.0,
    ):
        self.texture = texture

        self._scale = scale
        self._scale_transform = Mat3().scale(scale, scale)

        self._angle = angle
        self._angle_transform = Mat3().rotate(angle)

        self._offset = offset
        self._offset_transform = Mat3().translate(offset[0], offset[1])

    @property
    def pixel_transform(self):
        return self._offset_transform @ self._angle_transform @ self._scale_transform

    @property
    def scale(self) -> float:
        """
        Get or set the scale of the texture. This is a multiplier on the size of the texture.
        Default value is ``1.0``.
        """
        return self._scale

    @scale.setter
    def scale(self, value: float):
        self._scale = value
        self._scale_transform = Mat3().scale(value, value)

    @property
    def angle(self) -> float:
        """
        Get or set the angle of the texture. This is a rotation in radians.
        Default value is ``0.0``.
        """
        return self._angle

    @angle.setter
    def angle(self, value: float):
        self._angle = value
        self._angle_transform = Mat3().rotate(value)

    @property
    def offset(self) -> tuple[float, float]:
        """
        Get or set the offset of the texture. This is a translation in pixels.
        Default value is ``(0.0, 0.0)``.
        """
        return self._offset

    @offset.setter
    def offset(self, value: tuple[float, float]):
        self._offset = value
        self._offset_transform = Mat3().translate(-value[0], value[1])

    @property
    def wrap_x(self) -> int:
        """
        Get or set the horizontal wrapping of the texture. This decides how textures
        are read when texture coordinates are outside the ``[0.0, 1.0]`` area.
        Default value is ``REPEAT``.

        Valid options are::

            # Note: Enums can also be accessed in arcade.gl.
            # Repeat pixels on the y-axis.
            texture.wrap_x = ctx.REPEAT
            # Repeat pixels on the y-axis mirrored.
            texture.wrap_x = ctx.MIRRORED_REPEAT
            # Repeat the edge pixels when reading outside the texture.
            texture.wrap_x = ctx.CLAMP_TO_EDGE
            # Use the border color (black by default) when reading outside the texture.
            texture.wrap_x = ctx.CLAMP_TO_BORDER

        :type: int
        """
        return self.texture.wrap_x

    @wrap_x.setter
    def wrap_x(self, value: int):
        self.texture.wrap_x = value

    @property
    def wrap_y(self) -> int:
        """
        Get or set the horizontal wrapping of the texture. This decides how textures
        are read when texture coordinates are outside the ``[0.0, 1.0]`` area.
        Default value is ``REPEAT``.

        Valid options are::

            # Note: Enums can also be accessed in arcade.gl.
            # Repeat pixels on the y-axis.
            texture.wrap_y = ctx.REPEAT
            # Repeat pixels on the y-axis mirrored.
            texture.wrap_y = ctx.MIRRORED_REPEAT
            # Repeat the edge pixels when reading outside the texture.
            texture.wrap_y = ctx.CLAMP_TO_EDGE
            # Use the border color (black by default) when reading outside the texture.
            texture.wrap_y = ctx.CLAMP_TO_BORDER

        :type: int
        """
        return self.texture.wrap_y

    @wrap_y.setter
    def wrap_y(self, value: int):
        self.texture.wrap_y = value

    def use(self, unit: int = 0) -> None:
        """Bind the texture to a channel,

        Args:
            unit: The texture unit to bind the texture.
        """
        self.texture.use(unit)

    def render_target(
        self,
        context: ArcadeContext,
        color_attachments: list[gl.Texture2D] | None = None,
        depth_attachment: gl.Texture2D | None = None,
    ) -> gl.Framebuffer:
        """
        Create a framebuffer for the texture.

        This framebuffer is used to render to the texture. The framebuffer is created with the
        texture as the color attachment.

        Args:
            context: The context to use for the framebuffer.
            color_attachments: The color attachments to use for the framebuffer."
            depth_attachment: The depth attachment to use for the framebuffer."
        """
        if color_attachments is None:
            color_attachments = []
        return context.framebuffer(
            color_attachments=[self.texture] + color_attachments,
            depth_attachment=depth_attachment,
        )

    @staticmethod
    def from_file(
        tex_src: str,
        offset: tuple[float, float] = (0.0, 0.0),
        scale: float = 1.0,
        angle: float = 0.0,
        filters=(gl.NEAREST, gl.NEAREST),
    ):
        """ "
        Create a BackgroundTexture from a file.
        This is a convenience function to create a BackgroundTexture from a file.

        The file is loaded using PIL and converted to a texture.

        Args:
            tex_src: The file to load.
            offset: The offset of the texture in pixels.
            scale: The scale of the texture.
            angle: The angle of the texture in radians.
            filters: The filters to use for the texture.
        """
        _context = get_window().ctx

        with Image.open(resolve(tex_src)).convert("RGBA") as img:
            texture = _context.texture(
                img.size,
                data=img.transpose(Image.Transpose.FLIP_TOP_BOTTOM).tobytes(),
                filter=filters,
            )

        return BackgroundTexture(texture, offset, scale, angle)



import math

import arcade
from arcade.future.light import Light, LightLayer

# Do the math to figure out our screen dimensions
SCREEN_WIDTH = 1280
SCREEN_HEIGHT = 720
SCREEN_TITLE = "Lighting Demo (Experimental)"


class MyGame(arcade.Window):
    """
    Set up the application.
    """

    def __init__(self, width, height, title):
        super().__init__(width, height, title)
        self.background = arcade.load_texture(":resources:images/backgrounds/abstract_1.jpg")

        self.torch_list: arcade.SpriteList[arcade.Sprite] = arcade.SpriteList()
        self.torch_list.extend(
            [
                arcade.Sprite(
                    ":resources:images/tiles/torch1.png", scale=0.4, center_x=100, center_y=150
                ),
                arcade.Sprite(
                    ":resources:images/tiles/torch1.png", scale=0.4, center_x=300, center_y=150
                ),
                arcade.Sprite(
                    ":resources:images/tiles/torch1.png", scale=0.4, center_x=500, center_y=150
                ),
                arcade.Sprite(
                    ":resources:images/tiles/torch1.png", scale=0.4, center_x=700, center_y=150
                ),
                arcade.Sprite(
                    ":resources:images/tiles/torch1.png", scale=0.4, center_x=100, center_y=450
                ),
                arcade.Sprite(
                    ":resources:images/tiles/torch1.png", scale=0.4, center_x=300, center_y=450
                ),
                arcade.Sprite(
                    ":resources:images/tiles/torch1.png", scale=0.4, center_x=500, center_y=450
                ),
                arcade.Sprite(
                    ":resources:images/tiles/torch1.png", scale=0.4, center_x=700, center_y=450
                ),
            ]
        )

        self.light_layer = LightLayer(SCREEN_WIDTH, SCREEN_HEIGHT)
        # Add lights to the location of the torches. We're just using hacky tweak value list here
        params = (
            (100, "hard"),
            (100, "hard"),
            (100, "hard"),
            (100, "hard"),
            (120, "soft"),
            (120, "soft"),
            (120, "soft"),
            (120, "soft"),
        )
        for sprite, p in zip(self.torch_list, params):
            self.light_layer.add(
                Light(sprite.center_x, sprite.center_y, radius=p[0], mode=p[1]),
            )
        self.moving_light = Light(400, 300, radius=300, mode="soft")
        self.light_layer.add(self.moving_light)

    def on_draw(self):
        self.clear()

        # Everything that should be affected by lights in here
        with self.light_layer:
            arcade.draw_texture_rect(
                self.background, arcade.LBWH(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT)
            )
            self.torch_list.draw()

        # Draw the contents with lighting
        self.light_layer.draw()
        # draw with ambient
        # self.light_layer.draw(ambient_color=(127, 127, 127))

    def on_update(self, dt):
        # Keep track of elapsed time
        self.moving_light.position = (
            400 + math.sin(self.time) * 300,
            300 + math.cos(self.time) * 50,
        )
        self.moving_light.radius = 300 + math.sin(self.time * 2.34) * 150

    def on_resize(self, width, height):
        self.default_camera.use()
        self.light_layer.resize(width, height)


if __name__ == "__main__":
    MyGame(SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_TITLE)
    arcade.run()



"""
Show how to have enemies shoot bullets at random intervals.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_bullets_random
"""
import arcade
import random

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sprites and Random Bullets Example"


class GameView(arcade.View):
    """ Main application class """

    def __init__(self):
        super().__init__()

        self.background_color = arcade.color.BLACK

        self.frame_count = 0
        self.player_list = None
        self.enemy_list = None
        self.bullet_list = None

        self.player = None

    def setup(self):
        """ Setup the variables for the game. """
        self.player_list = arcade.SpriteList()
        self.enemy_list = arcade.SpriteList()
        self.bullet_list = arcade.SpriteList()

        # Add player ship
        self.player = arcade.Sprite(
            ":resources:images/space_shooter/playerShip1_orange.png",
            scale=0.5,
        )
        self.player_list.append(self.player)

        # Add top-left enemy ship
        enemy = arcade.Sprite(
            ":resources:images/space_shooter/playerShip1_green.png",
            scale=0.5,
        )
        enemy.center_x = 120
        enemy.center_y = WINDOW_HEIGHT - enemy.height
        enemy.angle = 180
        self.enemy_list.append(enemy)

        # Add top-right enemy ship
        enemy = arcade.Sprite(
            ":resources:images/space_shooter/playerShip1_green.png",
            scale=0.5,
        )
        enemy.center_x = WINDOW_WIDTH - 120
        enemy.center_y = WINDOW_HEIGHT - enemy.height
        enemy.angle = 180
        self.enemy_list.append(enemy)

    def on_draw(self):
        """Render the screen. """

        self.clear()

        self.enemy_list.draw()
        self.bullet_list.draw()
        self.player_list.draw()

    def on_update(self, delta_time):
        """All the logic to move, and the game logic goes here. """

        # Loop through each enemy that we have
        for enemy in self.enemy_list:

            # Have a random 1 in 200 change of shooting each 1/60th of a second
            odds = 200

            # Adjust odds based on delta-time
            adj_odds = int(odds * (1 / 60) / delta_time)

            if random.randrange(adj_odds) == 0:
                bullet = arcade.Sprite(":resources:images/space_shooter/laserBlue01.png")
                bullet.center_x = enemy.center_x
                bullet.angle = 90
                bullet.top = enemy.bottom
                bullet.change_y = -2
                self.bullet_list.append(bullet)

        # Get rid of the bullet when it flies off-screen
        for bullet in self.bullet_list:
            if bullet.top < 0:
                bullet.remove_from_sprite_lists()

        self.bullet_list.update()

    def on_mouse_motion(self, x, y, delta_x, delta_y):
        """ Called whenever the mouse moves. """
        self.player.center_x = x
        self.player.center_y = 20


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



import random

import arcade
from arcade.texture import transforms

TEST_TEXTURE_PATH = ":resources:images/test_textures/xy_square.png"
TRANSFORMS = [
    transforms.Rotate90Transform,
    transforms.Rotate180Transform,
    transforms.Rotate270Transform,
    transforms.FlipLeftRightTransform,
    transforms.FlipTopBottomTransform,
    transforms.TransposeTransform,
    transforms.TransverseTransform,
]


class App(arcade.Window):
    def __init__(self):
        super().__init__(1200, 600, "Atlas Revamp Check")
        paths = [
            ":resources:images/animated_characters/female_adventurer/femaleAdventurer_idle.png",
            ":resources:images/animated_characters/female_adventurer/femaleAdventurer_walk1.png",
            ":resources:images/animated_characters/female_adventurer/femaleAdventurer_walk2.png",
            ":resources:images/animated_characters/female_adventurer/femaleAdventurer_walk3.png",
            ":resources:images/animated_characters/female_adventurer/femaleAdventurer_walk4.png",
            ":resources:images/animated_characters/female_adventurer/femaleAdventurer_walk5.png",
            ":resources:images/animated_characters/female_adventurer/femaleAdventurer_walk6.png",
            ":resources:images/animated_characters/female_adventurer/femaleAdventurer_walk7.png",
        ]
        self.textures = [arcade.load_texture(path) for path in paths]
        self.spritelist = arcade.SpriteList()
        for i, tex in enumerate(self.textures):
            self.spritelist.append(arcade.Sprite(tex, center_x=100 + 130 * i, center_y=300))

        for i in range(len(TRANSFORMS)):
            sprite = self.spritelist[i]
            sprite.texture = sprite.texture.transform(TRANSFORMS[i])
            sprite.sync_hit_box_to_texture()

    def on_draw(self):
        self.clear()
        self.spritelist.draw()
        self.spritelist.draw_hit_boxes(color=arcade.color.RED, line_thickness=2)

    def on_key_press(self, symbol: int, modifiers: int):
        for sprite in self.spritelist:
            sprite.texture = sprite.texture.transform(random.choice(TRANSFORMS))
            sprite.sync_hit_box_to_texture()


App().run()



"""
Example showing a simple geometry shader.

We generate a bunch of random points and draw them using POINTS mode.
The geometry shader if configured to receive points and emit triangle strip.
We generate two triangles on the fly displaying a quad.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.geometry_shader
"""

import random
from array import array
import arcade
from arcade.gl import BufferDescription

# Do the math to figure out our screen dimensions
WINDOW_WIDTH = 800
WINDOW_HEIGHT = 800
WINDOW_TITLE = "Geometry Shader"


class GameView(arcade.Window):

    def __init__(self, width, height, title):
        """
        Set up the application.
        """
        super().__init__(width, height, title, resizable=True)
        self.program = self.ctx.program(
            vertex_shader="""
            #version 330
            in vec2 in_vert;
            void main() {
                gl_Position = vec4(in_vert, 0.0, 1.0);
            }
            """,
            geometry_shader="""
            #version 330

            layout (points) in;
            // Emit 4 vertices per point (Makes two triangles with a triangle strip)
            layout (triangle_strip, max_vertices = 4) out;

            uniform float time;
            out vec3 gs_color;

            void main() {
                // Get the position emitted from the vertex shader
                // We get the 0th element because point primitives only have 1 position.
                vec4 pos = gl_in[0].gl_Position;
                float primId = float(gl_PrimitiveIDIn);
                float offset = 0.02 + sin(time * 5.0 + primId) / 200.0;
                float rot = time + length(pos) * 10.0;

                mat2 rotate = mat2(
                    cos(rot), -sin(rot),
                    sin(rot),  cos(rot)
                );
                vec3 color = vec3(
                    (sin(time + primId) + 1.0) / 2.0,
                    (sin(time + 2.0 + primId) + 1.0) / 2.0,
                    (sin(time + 3.0 + primId) + 1.0) / 2.0
                );
                // Emit 4 vertices around the original vertex position making a quad
                gl_Position = pos + vec4(rotate * vec2(-offset, offset), 0.0, 0.0);
                gs_color = color;
                EmitVertex();

                gl_Position = pos + vec4(rotate * vec2(-offset, -offset), 0.0, 0.0);
                gs_color = color;
                EmitVertex();

                gl_Position = pos + vec4(rotate * vec2(offset, offset), 0.0, 0.0);
                gs_color = color;
                EmitVertex();

                gl_Position = pos + vec4(rotate * vec2(offset, -offset), 0.0, 0.0);
                gs_color = color;
                EmitVertex();

                // Done with the primitive
                EndPrimitive();
            }
            """,
            fragment_shader="""
            #version 330
            out vec4 out_color;
            in vec3 gs_color;
            void main() {
                out_color = vec4(gs_color, 1.0);
            }
            """,
        )
        num_points = 1000
        self.points = self.ctx.geometry(
            [
                BufferDescription(
                    self.ctx.buffer(
                        data=array("f", [random.uniform(-1.0, 1.0) for _ in range(num_points * 2)])
                    ),
                    "2f",
                    ["in_vert"],
                )
            ]
        )

    def on_draw(self):
        self.clear()
        self.ctx.enable(self.ctx.BLEND)
        self.program["time"] = self.time
        self.points.render(self.program, mode=self.ctx.POINTS)


if __name__ == "__main__":
    GameView(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)
    arcade.run()



import math
import random

import arcade
from arcade.future.light import Light, LightLayer

# Do the math to figure out our screen dimensions
SCREEN_WIDTH = 1280
SCREEN_HEIGHT = 720
SCREEN_TITLE = "Lighting Demo (Experimental)"


class MyGame(arcade.Window):
    """
    Set up the application.
    """

    def __init__(self, width, height, title):
        super().__init__(width, height, title)
        self.background = arcade.load_texture(":resources:images/backgrounds/abstract_1.jpg")

        self.light_layer = LightLayer(SCREEN_WIDTH, SCREEN_HEIGHT)
        self.light_layer.set_background_color(arcade.color.WHITE)
        # Add some random lights
        for _ in range(500):
            self.light_layer.add(
                Light(
                    random.randrange(0, SCREEN_WIDTH),
                    random.randrange(0, SCREEN_HEIGHT),
                    radius=50,
                    mode="soft",
                    color=(
                        random.randint(0, 255),
                        random.randint(0, 255),
                        random.randint(0, 255),
                    ),
                ),
            )

    def on_draw(self):
        self.clear()

        # Everything that should be affected by lights in here
        with self.light_layer:
            # The light layer is just cleared using the background color here (white)
            pass

        # Draw the contents with lighting
        self.light_layer.draw()

        # image = arcade.get_image()
        # image.save(f'screenshots/frame{str(self.frame).zfill(3)}.png', 'png')

    def on_update(self, dt):
        # dt = 0.1
        try:
            for i, light in enumerate(self.light_layer):
                light.radius = 20 + math.sin(self.time + i) * 40
        except Exception as e:
            print(e)

    def on_resize(self, width, height):
        self.default_camera.use()
        self.light_layer.resize(width, height)


if __name__ == "__main__":
    MyGame(SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_TITLE)
    arcade.run()



from __future__ import annotations

import weakref
from contextlib import contextmanager
from ctypes import Array, c_int, c_uint, string_at
from typing import TYPE_CHECKING, Generator

from pyglet import gl

from arcade.types import RGBOrA255, RGBOrANormalized

from .texture import Texture2D
from .types import pixel_formats

if TYPE_CHECKING:
    from arcade.gl import Context


class Framebuffer:
    """
    An offscreen render target also called a Framebuffer Object in OpenGL.
    This implementation is using texture attachments. When creating a
    Framebuffer we supply it with textures we want our scene rendered into.
    The advantage of using texture attachments is the ability we get
    to keep working on the contents of the framebuffer.

    The best way to create framebuffer is through :py:meth:`arcade.gl.Context.framebuffer`::

        # Create a 100 x 100 framebuffer with one attachment
        ctx.framebuffer(color_attachments=[ctx.texture((100, 100), components=4)])

        # Create a 100 x 100 framebuffer with two attachments
        # Shaders can be configured writing to the different layers
        ctx.framebuffer(
            color_attachments=[
                ctx.texture((100, 100), components=4),
                ctx.texture((100, 100), components=4),
            ]
        )

    Args:
        ctx:
            The context this framebuffer belongs to
        color_attachments:
            A color attachment or a list of color attachments
        depth_attachment:
            A depth attachment
    """

    #: Is this the default framebuffer? (window buffer)
    is_default = False
    __slots__ = (
        "_ctx",
        "_glo",
        "_width",
        "_height",
        "_color_attachments",
        "_depth_attachment",
        "_samples",
        "_viewport",
        "_scissor",
        "_depth_mask",
        "_draw_buffers",
        "_prev_fbo",
        "__weakref__",
    )

    def __init__(
        self,
        ctx: "Context",
        *,
        color_attachments: Texture2D | list[Texture2D],
        depth_attachment: Texture2D | None = None,
    ):
        self._glo = fbo_id = gl.GLuint()  # The OpenGL alias/name
        self._ctx = ctx
        if not color_attachments:
            raise ValueError("Framebuffer must at least have one color attachment")

        self._color_attachments = (
            color_attachments if isinstance(color_attachments, list) else [color_attachments]
        )
        self._depth_attachment: Texture2D | None = depth_attachment
        self._samples = 0  # Leaving this at 0 for future sample support
        self._depth_mask = True  # Determines if the depth buffer should be affected
        self._prev_fbo = None

        # Create the framebuffer object
        gl.glGenFramebuffers(1, self._glo)
        gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, self._glo)

        # Ensure all attachments have the same size.
        # OpenGL do actually support different sizes,
        # but let's keep this simple with high compatibility.
        self._width, self._height = self._detect_size()
        self._viewport = 0, 0, self._width, self._height
        self._scissor: tuple[int, int, int, int] | None = None

        # Attach textures to it
        for i, tex in enumerate(self._color_attachments):
            # TODO: Possibly support attaching a specific mipmap level
            #       but we can read from specific mip levels from shaders.
            gl.glFramebufferTexture2D(
                gl.GL_FRAMEBUFFER,
                gl.GL_COLOR_ATTACHMENT0 + i,
                tex._target,
                tex.glo,
                0,  # Level 0
            )

        if self.depth_attachment:
            gl.glFramebufferTexture2D(
                gl.GL_FRAMEBUFFER,
                gl.GL_DEPTH_ATTACHMENT,
                self.depth_attachment._target,
                self.depth_attachment.glo,
                0,
            )

        # Ensure the framebuffer is sane!
        self._check_completeness()

        # Set up draw buffers. This is simply a prepared list of attachments enums
        # we use in the use() method to activate the different color attachment layers
        layers = [gl.GL_COLOR_ATTACHMENT0 + i for i, _ in enumerate(self._color_attachments)]
        # pyglet wants this as a ctypes thingy, so let's prepare it
        self._draw_buffers: Array[c_uint] | None = (gl.GLuint * len(layers))(*layers)

        # Restore the original bound framebuffer to avoid confusion
        self.ctx.active_framebuffer.use(force=True)

        if self._ctx.gc_mode == "auto" and not self.is_default:
            weakref.finalize(self, Framebuffer.delete_glo, ctx, fbo_id)

        self.ctx.stats.incr("framebuffer")

    def __del__(self):
        # Intercept garbage collection if we are using Context.gc()
        if self._ctx.gc_mode == "context_gc" and not self.is_default and self._glo.value > 0:
            self._ctx.objects.append(self)

    @property
    def glo(self) -> gl.GLuint:
        """The OpenGL id/name of the framebuffer."""
        return self._glo

    def _get_viewport(self) -> tuple[int, int, int, int]:
        """
        Get or set the framebuffer's viewport.

        The viewport parameter are ``(x, y, width, height)``.
        It determines what part of the framebuffer should be rendered to.
        By default the viewport is ``(0, 0, width, height)``.

        The viewport value is persistent all will automatically
        be applies every time the framebuffer is bound.

        Example::

            # 100, x 100 lower left with size 200 x 200px
            fb.viewport = 100, 100, 200, 200
        """
        return self._viewport

    def _set_viewport(self, value: tuple[int, int, int, int]):
        if not isinstance(value, tuple) or len(value) != 4:
            raise ValueError("viewport should be a 4-component tuple")

        self._viewport = value

        # If the framebuffer is bound we need to set the viewport.
        # Otherwise it will be set on use()
        if self._ctx.active_framebuffer == self:
            gl.glViewport(*self._viewport)
            if self._scissor is None:
                gl.glScissor(*self._viewport)
            else:
                gl.glScissor(*self._scissor)

    viewport = property(_get_viewport, _set_viewport)

    def _get_scissor(self) -> tuple[int, int, int, int] | None:
        """
        Get or set the scissor box for this framebuffer.

        By default the scissor box is disabled and has no effect
        and will have an initial value of ``None``. The scissor
        box is enabled when setting a value and disabled when
        set to ``None``

            # Set and enable scissor box only drawing
            # in a 100 x 100 pixel lower left area
            ctx.scissor = 0, 0, 100, 100
            # Disable scissoring
            ctx.scissor = None

        :type: tuple (x, y, width, height)

        """
        return self._scissor

    def _set_scissor(self, value):
        self._scissor = value

        if self._scissor is None:
            if self._ctx.active_framebuffer == self:
                gl.glScissor(*self._viewport)
        else:
            if self._ctx.active_framebuffer == self:
                gl.glScissor(*self._scissor)

    scissor = property(_get_scissor, _set_scissor)

    @property
    def ctx(self) -> "Context":
        """The context this object belongs to."""
        return self._ctx

    @property
    def width(self) -> int:
        """The width of the framebuffer in pixels."""
        return self._width

    @property
    def height(self) -> int:
        """The height of the framebuffer in pixels."""
        return self._height

    @property
    def size(self) -> tuple[int, int]:
        """Size as a ``(w, h)`` tuple."""
        return self._width, self._height

    @property
    def samples(self) -> int:
        """Number of samples (MSAA)."""
        return self._samples

    @property
    def color_attachments(self) -> list[Texture2D]:
        """A list of color attachments."""
        return self._color_attachments

    @property
    def depth_attachment(self) -> Texture2D | None:
        """Depth attachment."""
        return self._depth_attachment

    @property
    def depth_mask(self) -> bool:
        """
        Get or set the depth mask (default: ``True``).

        It determines if depth values should be written
        to the depth texture when depth testing is enabled.

        The depth mask value is persistent all will automatically
        be applies every time the framebuffer is bound.
        """
        return self._depth_mask

    @depth_mask.setter
    def depth_mask(self, value: bool):
        self._depth_mask = value
        # Set state if framebuffer is active
        if self._ctx.active_framebuffer == self:
            gl.glDepthMask(self._depth_mask)

    def __enter__(self):
        self._prev_fbo = self._ctx.active_framebuffer
        self.use()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._prev_fbo:
            self._prev_fbo.use()

    @contextmanager
    def activate(self) -> Generator[Framebuffer, None, None]:
        """Context manager for binding the framebuffer.

        Unlike the default context manager in this class
        this support nested framebuffer binding.
        """
        prev_fbo = self._ctx.active_framebuffer
        try:
            self.use()
            yield self
        finally:
            prev_fbo.use()

    def use(self, *, force: bool = False):
        """Bind the framebuffer making it the target of all rendering commands

        Args:
            force:
                Force the framebuffer binding even if the system
                already believes it's already bound.
        """
        self._use(force=force)
        self._ctx.active_framebuffer = self

    def _use(self, *, force: bool = False):
        """Internal use that do not change the global active framebuffer"""
        if self.ctx.active_framebuffer == self and not force:
            return

        gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, self._glo)

        # NOTE: gl.glDrawBuffer(GL_NONE) if no texture attachments (future)
        # NOTE: Default framebuffer currently has this set to None
        if self._draw_buffers:
            gl.glDrawBuffers(len(self._draw_buffers), self._draw_buffers)

        gl.glDepthMask(self._depth_mask)
        gl.glViewport(*self._viewport)
        if self._scissor is not None:
            gl.glScissor(*self._scissor)
        else:
            gl.glScissor(*self._viewport)

    def clear(
        self,
        *,
        color: RGBOrA255 | None = None,
        color_normalized: RGBOrANormalized | None = None,
        depth: float = 1.0,
        viewport: tuple[int, int, int, int] | None = None,
    ):
        """
        Clears the framebuffer::

            # Clear the framebuffer using Arcade's colors (not normalized)
            fb.clear(color=arcade.color.WHITE)

            # Clear framebuffer using the color red in normalized form
            fbo.clear(color_normalized=(1.0, 0.0, 0.0, 1.0))

        If the background color is an ``RGB`` value instead of ``RGBA```
        we assume alpha value 255.

        Args:
            color:
                A 3 or 4 component tuple containing the color
                (prioritized over color_normalized)
            color_normalized:
                A 3 or 4 component tuple containing the color in normalized form
            depth:
                Value to clear the depth buffer (unused)
            viewport:
                The viewport range to clear
        """
        with self.activate():
            scissor_values = self._scissor

            if viewport:
                self.scissor = viewport
            else:
                self.scissor = None

            clear_color = 0.0, 0.0, 0.0, 0.0
            if color is not None:
                if len(color) == 3:
                    clear_color = color[0] / 255, color[1] / 255, color[2] / 255, 1.0
                elif len(color) == 4:
                    clear_color = color[0] / 255, color[1] / 255, color[2] / 255, color[3] / 255
                else:
                    raise ValueError("Color should be a 3 or 4 component tuple")
            elif color_normalized is not None:
                if len(color_normalized) == 3:
                    clear_color = color_normalized[0], color_normalized[1], color_normalized[2], 1.0
                elif len(color_normalized) == 4:
                    clear_color = color_normalized
                else:
                    raise ValueError("Color should be a 3 or 4 component tuple")

            gl.glClearColor(*clear_color)

            if self.depth_attachment:
                if self._ctx.gl_api == "gl":
                    gl.glClearDepth(depth)
                else:  # gles only supports glClearDepthf
                    gl.glClearDepthf(depth)

                gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
            else:
                gl.glClear(gl.GL_COLOR_BUFFER_BIT)

            self.scissor = scissor_values

    def read(self, *, viewport=None, components=3, attachment=0, dtype="f1") -> bytes:
        """
        Read the raw framebuffer pixels.

        Reading data from a framebuffer is much more powerful than
        reading date from textures. We can specify more or less
        what format we want the data. It's not uncommon to throw
        textures into framebuffers just to get access to this read
        api.

        Args:
            viewport:
                The x, y, with, height area to read.
            components:
                The number of components to read. 1, 2, 3 or 4.
                This will determine the format to read.
            attachment:
                The attachment id to read from
            dtype:
                The data type to read. Pixel data will be converted to this format.
        """
        # TODO: use texture attachment info to determine read format?
        try:
            frmt = pixel_formats[dtype]
            base_format = frmt[0][components]
            pixel_type = frmt[2]
            component_size = frmt[3]
        except Exception:
            raise ValueError(f"Invalid dtype '{dtype}'")

        with self.activate():
            # Configure attachment to read from. Does not work on default framebuffer.
            if not self.is_default:
                gl.glReadBuffer(gl.GL_COLOR_ATTACHMENT0 + attachment)

            gl.glPixelStorei(gl.GL_PACK_ALIGNMENT, 1)
            gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 1)

            if viewport:
                x, y, width, height = viewport
            else:
                x, y, width, height = 0, 0, *self.size

            data = (gl.GLubyte * (components * component_size * width * height))(0)
            gl.glReadPixels(x, y, width, height, base_format, pixel_type, data)

            if not self.is_default:
                gl.glReadBuffer(gl.GL_COLOR_ATTACHMENT0)  # Reset to default

        return string_at(data, len(data))

    def resize(self):
        """
        Detects size changes in attachments.

        This will reset the viewport to ``0, 0, width, height``.
        """
        self._width, self._height = self._detect_size()
        self.viewport = 0, 0, self.width, self._height

    def delete(self):
        """
        Destroy the underlying OpenGL resource.

        .. warning:: Don't use this unless you know exactly what you are doing.
        """
        Framebuffer.delete_glo(self._ctx, self._glo)
        self._glo.value = 0

    @staticmethod
    def delete_glo(ctx, framebuffer_id):
        """
        Destroys the framebuffer object

        Args:
            ctx:
                The context this framebuffer belongs to
            framebuffer_id:
                Framebuffer id destroy (glo)
        """
        if gl.current_context is None:
            return

        gl.glDeleteFramebuffers(1, framebuffer_id)
        ctx.stats.decr("framebuffer")

    def _detect_size(self) -> tuple[int, int]:
        """Detect the size of the framebuffer based on the attachments"""
        attachments = []

        if self._color_attachments:
            attachments.extend(self._color_attachments)
        elif self._depth_attachment:
            attachments.append(self._depth_attachment)

        if not attachments:
            raise RuntimeError("Framebuffer has no attachments")

        expected_size = attachments[0].size
        for layer in attachments:
            if layer and layer.size != expected_size:
                raise ValueError("All framebuffer attachments should have the same size")
        return expected_size

    @staticmethod
    def _check_completeness() -> None:
        """
        Checks the completeness of the framebuffer.

        If the framebuffer is not complete, we cannot continue.
        """
        # See completeness rules : https://www.khronos.org/opengl/wiki/Framebuffer_Object
        states = {
            gl.GL_FRAMEBUFFER_UNSUPPORTED: "Framebuffer unsupported. Try another format.",
            gl.GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT: "Framebuffer incomplete attachment.",
            gl.GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT: "Framebuffer missing attachment.",
            gl.GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT: "Framebuffer unsupported dimension.",
            gl.GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT: "Framebuffer incomplete formats.",
            gl.GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER: "Framebuffer incomplete draw buffer.",
            gl.GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER: "Framebuffer incomplete read buffer.",
            gl.GL_FRAMEBUFFER_COMPLETE: "Framebuffer is complete.",
        }

        status = gl.glCheckFramebufferStatus(gl.GL_FRAMEBUFFER)
        if status != gl.GL_FRAMEBUFFER_COMPLETE:
            raise ValueError(
                "Framebuffer is incomplete. {}".format(states.get(status, "Unknown error"))
            )

    def __repr__(self):
        return "<Framebuffer glo={}>".format(self._glo.value)


class DefaultFrameBuffer(Framebuffer):
    """
    Represents the default framebuffer.

    This is the framebuffer of the window itself and need some special handling.

    We are not allowed to destroy this framebuffer since it's owned by pyglet.
    This framebuffer can also change size and pixel ratio at any point.

    We're doing some initial introspection to guess somewhat sane initial values.
    Since this is a dynamic framebuffer we cannot trust the internal values.
    We can only trust what the pyglet window itself reports related to window size
    and framebuffer size. This should be updated in the ``on_resize`` callback.
    """

    is_default = True
    """Is this the default framebuffer? (window buffer)"""

    __slots__ = ()

    def __init__(self, ctx: "Context"):
        self._ctx = ctx
        # TODO: Can we query this?
        self._samples = 0
        # TODO: Maybe we should map renderbuffers?
        self._color_attachments = []
        self._depth_attachment = None
        self._depth_mask = True

        value = c_int()
        gl.glGetIntegerv(gl.GL_DRAW_FRAMEBUFFER_BINDING, value)
        self._glo = gl.GLuint(value.value)

        # Query draw buffers. This will most likely return GL_BACK
        # value = c_int()
        # gl.glGetIntegerv(gl.GL_DRAW_BUFFER0, value)
        # print(value.value)  # 1029 GL_BACK 0x405
        # self._draw_buffers = (gl.GLuint * 1)(value.value)
        # NOTE: Don't query for now
        self._draw_buffers = None

        # Query viewport values by inspecting the scissor box
        values = (c_int * 4)()
        gl.glGetIntegerv(gl.GL_SCISSOR_BOX, values)
        x, y, width, height = list(values)

        self._viewport = x, y, width, height
        self._scissor = None
        self._width = width
        self._height = height

        # HACK: Signal the default framebuffer having depth buffer
        self._depth_attachment = True  # type: ignore

    @property
    def size(self) -> tuple[int, int]:
        """Size as a ``(w, h)`` tuple."""
        return self._ctx.window.get_framebuffer_size()

    @property
    def width(self) -> int:
        """The width of the framebuffer in pixels."""
        return self.size[0]

    @property
    def height(self) -> int:
        """The height of the framebuffer in pixels."""
        return self.size[1]

    def _get_framebuffer_size(self) -> tuple[int, int]:
        """Get the framebuffer size of the window"""
        return self._ctx.window.get_framebuffer_size()

    def _get_viewport(self) -> tuple[int, int, int, int]:
        """
        Get or set the framebuffer's viewport.
        The viewport parameter are ``(x, y, width, height)``.
        It determines what part of the framebuffer should be rendered to.
        By default the viewport is ``(0, 0, width, height)``.

        The viewport value is persistent all will automatically
        be applies every time the framebuffer is bound.

        Example::

            # 100, x 100 lower left with size 200 x 200px
            fb.viewport = 100, 100, 200, 200
        """
        ratio = self.ctx.window.get_pixel_ratio()
        return (
            int(self._viewport[0] / ratio),
            int(self._viewport[1] / ratio),
            int(self._viewport[2] / ratio),
            int(self._viewport[3] / ratio),
        )

    def _set_viewport(self, value: tuple[int, int, int, int]):
        if not isinstance(value, tuple) or len(value) != 4:
            raise ValueError("viewport should be a 4-component tuple")

        ratio = self.ctx.window.get_pixel_ratio()
        self._viewport = (
            int(value[0] * ratio),
            int(value[1] * ratio),
            int(value[2] * ratio),
            int(value[3] * ratio),
        )

        # If the framebuffer is bound we need to set the viewport.
        # Otherwise it will be set on use()
        if self._ctx.active_framebuffer == self:
            gl.glViewport(*self._viewport)
            if self._scissor is None:
                # FIXME: Probably should be set to the framebuffer size
                gl.glScissor(*self._viewport)
            else:
                gl.glScissor(*self._scissor)

    viewport = property(_get_viewport, _set_viewport)

    def _get_scissor(self) -> tuple[int, int, int, int] | None:
        """
        Get or set the scissor box for this framebuffer.

        By default the scissor box is disabled and has no effect
        and will have an initial value of ``None``. The scissor
        box is enabled when setting a value and disabled when
        set to ``None``::

            # Set and enable scissor box only drawing
            # in a 100 x 100 pixel lower left area
            ctx.scissor = 0, 0, 100, 100
            # Disable scissoring
            ctx.scissor = None
        """
        if self._scissor is None:
            return None

        ratio = self.ctx.window.get_pixel_ratio()
        return (
            int(self._scissor[0] / ratio),
            int(self._scissor[1] / ratio),
            int(self._scissor[2] / ratio),
            int(self._scissor[3] / ratio),
        )

    def _set_scissor(self, value):
        if value is None:
            # FIXME: Do we need to reset something here?
            self._scissor = None
            if self._ctx.active_framebuffer == self:
                gl.glScissor(*self._viewport)
        else:
            ratio = self.ctx.window.get_pixel_ratio()
            self._scissor = (
                int(value[0] * ratio),
                int(value[1] * ratio),
                int(value[2] * ratio),
                int(value[3] * ratio),
            )

            # If the framebuffer is bound we need to set the scissor box.
            # Otherwise it will be set on use()
            if self._ctx.active_framebuffer == self:
                gl.glScissor(*self._scissor)

    scissor = property(_get_scissor, _set_scissor)



"""
Sprite Explosion

Simple program to show how to make explosions with a series of bitmaps.

Artwork from https://kenney.nl
Explosion graphics from https://www.explosiongenerator.com/

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_explosion_bitmapped
"""
import random
import arcade

SPRITE_SCALING_PLAYER = 0.5
SPRITE_SCALING_COIN = 0.2
SPRITE_SCALING_LASER = 0.8
COIN_COUNT = 50

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sprite Explosion Example"

BULLET_SPEED = 5

EXPLOSION_TEXTURE_COUNT = 60


class Explosion(arcade.Sprite):
    """ This class creates an explosion animation """

    def __init__(self, texture_list):
        super().__init__(texture_list[0])
        # How long the explosion has been around.
        self.time_elapsed = 0

        # Start at the first frame
        self.current_texture = 0
        self.textures = texture_list

    def update(self, delta_time=1 / 60):
        self.time_elapsed += delta_time
        # Update to the next frame of the animation. If we are at the end
        # of our frames, then delete this sprite.
        self.current_texture = int(self.time_elapsed * 60)
        if self.current_texture < len(self.textures):
            self.set_texture(self.current_texture)
        else:
            self.remove_from_sprite_lists()


class GameView(arcade.View):
    """ Main application class. """

    def __init__(self):
        """ Initializer """
        # Call the parent class initializer
        super().__init__()

        # Variables that will hold sprite lists
        self.player_list = arcade.SpriteList()
        self.enemy_list = arcade.SpriteList()
        self.bullet_list = arcade.SpriteList()
        self.explosions_list = arcade.SpriteList()

        # Set up the player info
        # Image from kenney.nl
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=SPRITE_SCALING_PLAYER,
        )
        self.player_sprite.center_x = 50
        self.player_sprite.center_y = 70
        self.player_list.append(self.player_sprite)

        # Player score
        self.score = 0

        # Don't show the mouse cursor
        self.window.set_mouse_visible(False)

        # Pre-load the animation frames. We don't do this in the __init__
        # of the explosion sprite because it
        # takes too long and would cause the game to pause.
        self.explosion_texture_list = []

        # Load the explosion from a sprite sheet
        columns = 16
        count = 60
        sprite_width = 256
        sprite_height = 256
        file_name = ":resources:images/spritesheets/explosion.png"

        # Load the explosions from a sprite sheet
        spritesheet = arcade.load_spritesheet(file_name)
        self.explosion_texture_list = spritesheet.get_texture_grid(
            size=(sprite_width, sprite_height),
            columns=columns,
            count=count,
        )

        # Load sounds. Sounds from kenney.nl
        self.gun_sound = arcade.sound.load_sound(":resources:sounds/laser2.wav")
        self.hit_sound = arcade.sound.load_sound(":resources:sounds/explosion2.wav")

        self.background_color = arcade.color.AMAZON
        self.spawn_enemies()

    def reset(self):
        """Restart the game."""

        # Clear out the sprite lists
        self.enemy_list.clear()
        self.bullet_list.clear()
        self.explosions_list.clear()

        # Reset the score
        self.score = 0

        self.spawn_enemies()

    def spawn_enemies(self):
        for coin_index in range(COIN_COUNT):
            # Create the enemy instance. Image from kenney.nl
            coin = arcade.Sprite(
                ":resources:images/items/coinGold.png",
                scale=SPRITE_SCALING_COIN,
                center_x=random.randrange(25, WINDOW_WIDTH - 25),
                center_y=random.randrange(150, WINDOW_HEIGHT),
            )
            # Add the coin to enemy list
            self.enemy_list.append(coin)

    def on_draw(self):
        """
        Render the screen.
        """
        # This command has to happen before we start drawing
        self.clear()

        # Draw all the sprites.
        self.enemy_list.draw()
        self.bullet_list.draw()
        self.player_list.draw()
        self.explosions_list.draw()

        # Render the text
        arcade.draw_text(f"Score: {self.score}", 10, 20, arcade.color.WHITE, 14)

    def on_mouse_motion(self, x, y, dx, dy):
        """
        Called whenever the mouse moves.
        """
        self.player_sprite.center_x = x

    def on_mouse_press(self, x, y, button, modifiers):
        """
        Called whenever the mouse button is clicked.
        """
        # Gunshot sound
        arcade.sound.play_sound(self.gun_sound)

        # Create a bullet
        bullet = arcade.Sprite(
            ":resources:images/space_shooter/laserBlue01.png",
            scale=SPRITE_SCALING_LASER,
        )

        # The image points to the right, and we want it to point up. So
        # rotate it.
        bullet.angle = 270

        # Give it a speed
        bullet.change_y = BULLET_SPEED

        # Position the bullet
        bullet.center_x = self.player_sprite.center_x
        bullet.bottom = self.player_sprite.top

        # Add the bullet to the appropriate lists
        self.bullet_list.append(bullet)

    def on_key_press(self, symbol: int, modifiers: int):
        if symbol == arcade.key.R:
            self.reset()
        # Close the window
        elif symbol == arcade.key.ESCAPE:
            self.window.close()

    def on_update(self, delta_time):
        """Movement and game logic"""

        # Call update on bullet sprites
        self.bullet_list.update()
        self.explosions_list.update()

        # Loop through each bullet
        for bullet in self.bullet_list:

            # Check this bullet to see if it hit a coin
            hit_list = arcade.check_for_collision_with_list(bullet, self.enemy_list)

            # If it did...
            if len(hit_list) > 0:

                # Make an explosion
                explosion = Explosion(self.explosion_texture_list)

                # Move it to the location of the coin
                explosion.center_x = hit_list[0].center_x
                explosion.center_y = hit_list[0].center_y

                # Call update() because it sets which image we start on
                explosion.update()

                # Add to a list of sprites that are explosions
                self.explosions_list.append(explosion)

                # Get rid of the bullet
                bullet.remove_from_sprite_lists()

            # For every coin we hit, add to the score and remove the coin
            for coin in hit_list:
                coin.remove_from_sprite_lists()
                self.score += 1

                # Hit Sound
                arcade.sound.play_sound(self.hit_sound)

            # If the bullet flies off-screen, remove it.
            if bullet.bottom > WINDOW_HEIGHT:
                bullet.remove_from_sprite_lists()



def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()

if __name__ == "__main__":
    main()



"""
Area grips act similarly to position constrains used by physics engines. Much like
rotations the order of operations can have large impacts on the behavior of the
camera, so try and keep the number of area grips low.

The methods don't update the camera data directly incase you want to smoothly
interpolate towards the target position
"""

from arcade.camera.data_types import CameraData
from arcade.math import clamp
from arcade.types import Box, Point2, Point3, Rect


def constrain_x(data: CameraData, minimum: float, maximum: float) -> Point3:
    """
    Prevent the provided CameraData from leaving the provided bounds along the x-axis.
    assumed minimum < maximum
    """
    x, y, z = data.position
    if minimum <= x <= maximum:
        return data.position

    n_x = clamp(x, minimum, maximum)
    return n_x, y, z


def constrain_y(data: CameraData, minimum: float, maximum: float) -> Point3:
    """
    Prevent the provided CameraData from leaving the provided bounds along the y-axis.
    assumed minimum < maximum
    """
    x, y, z = data.position
    if minimum <= y <= maximum:
        return data.position

    n_y = clamp(y, minimum, maximum)
    return x, n_y, z


def constrain_z(data: CameraData, minimum: float, maximum: float) -> Point3:
    """
    Prevent the provided CameraData from leaving the provided bounds along the z-axis.
    assumed minimum < maximum
    """
    x, y, z = data.position
    if minimum <= z <= maximum:
        return data.position

    n_z = clamp(z, minimum, maximum)
    return x, y, n_z


def constrain_xy(data: CameraData, area: Rect) -> Point3:
    """
    Prevent the provided CameraData from leaving the provided Rect along the x and y-axis.
    Assumes area.left < area.right and area.bottom < area.top
    """
    x, y, z = data.position
    if area.point_in_rect((x, y)):
        return data.position

    n_x = clamp(x, area.left, area.right)
    n_y = clamp(y, area.bottom, area.top)

    return n_x, n_y, z


def constrain_yz(data: CameraData, area: Rect) -> Point3:
    """
    Prevent the provided CameraData from leaving the provided Rect along the y and z-axis.
    Assumes area.left < area.right and area.bottom < area.top
    """
    x, y, z = data.position
    if area.point_in_rect((y, z)):
        return data.position

    n_y = clamp(y, area.left, area.right)
    n_z = clamp(z, area.bottom, area.top)

    return x, n_y, n_z


def constrain_xz(data: CameraData, area: Rect) -> Point3:
    """
    Prevent the provided CameraData from leaving the provided Rect along the x and z-axis.
    Assumes area.left < area.right and area.bottom < area.top
    """
    x, y, z = data.position
    if area.point_in_rect((x, z)):
        return data.position

    n_x = clamp(x, area.left, area.right)
    n_z = clamp(z, area.bottom, area.top)

    return n_x, y, n_z


def constrain_xyz(data: CameraData, volume: Box) -> Point3:
    """
    Prevent the provided CameraData from leaving the provided Box along the x, y, and z-axis.
    assumes volume.left < volume.right, volume.bottom < volume.top, and volume.near < volume.far
    """
    x, y, z = data.position
    if volume.point_in_box((x, y, z)):
        return data.position

    n_x = clamp(x, volume.left, volume.right)
    n_y = clamp(y, volume.bottom, volume.top)
    n_z = clamp(z, volume.near, volume.far)

    return n_x, n_y, n_z


def constrain_boundary_x(data: CameraData, left: float, right: float, target_x: float) -> Point3:
    """
    Ensure that the target point is within the boundary values along the x-axis

    Assumes left < right
    """
    x, y, z = data.position
    dx = target_x - x
    if left <= dx <= right:
        return data.position

    n_x = target_x - clamp(dx, left, right)
    return n_x, y, z


def constrain_boundary_y(data: CameraData, bottom: float, top: float, target_y: float) -> Point3:
    """
    Ensure that the target point is within the boundary values along the y-axis

    Assumes bottom < top
    """
    x, y, z = data.position
    dy = target_y - y
    if bottom <= dy <= top:
        return data.position

    n_y = target_y - clamp(dy, bottom, top)
    return x, n_y, z


def constrain_boundary_z(data: CameraData, left: float, right: float, target_z: float) -> Point3:
    """
    Ensure that the target point is within the boundary values along the z-axis

    Assumes near < far
    """
    x, y, z = data.position
    dz = target_z - z
    if left <= dz <= right:
        return data.position

    n_z = target_z - clamp(dz, left, right)
    return x, y, n_z


def constrain_boundary_xy(data: CameraData, bounds: Rect, target: Point2) -> Point3:
    """
    Ensure that  the target point is within the boundary Rect along the x and y-axis

    Assumes bounds.left < bounds.right and bounds.bottom < bounds.top
    """
    x, y, z = data.position
    t_x, t_y = target
    dx = t_x - x
    dy = t_y - y
    if bounds.point_in_rect((dx, dy)):
        return data.position

    n_x = t_x - clamp(dx, bounds.left, bounds.right)
    n_y = t_y - clamp(dy, bounds.bottom, bounds.top)

    return n_x, n_y, z


def constrain_boundary_yz(data: CameraData, bounds: Rect, target: Point2) -> Point3:
    """
    Ensure that the target point is within the boundary Rect along the y and z-axis

    Assumes bounds.left < bounds.right and bounds.bottom < bounds.top
    """
    x, y, z = data.position
    t_y, t_z = target
    dy = t_y - y
    dz = t_z - z
    if bounds.point_in_rect((dy, dz)):
        return data.position

    n_y = t_y - clamp(dy, bounds.left, bounds.right)
    n_z = t_z - clamp(dz, bounds.bottom, bounds.top)

    return x, n_y, n_z


def constrain_boundary_xz(data: CameraData, bounds: Rect, target: Point2) -> Point3:
    """
    Ensure that the target point is within the boundary Rect along the x and z-axis

    Assumes bounds.left < bounds.right and bounds.bottom < bounds.top
    """
    x, y, z = data.position
    t_x, t_z = target
    dx = t_x - x
    dz = t_z - z
    if bounds.point_in_rect((dx, dz)):
        return data.position

    n_x = t_x - clamp(dx, bounds.left, bounds.right)
    n_z = t_z - clamp(dz, bounds.bottom, bounds.top)

    return n_x, y, n_z


def constrain_boundary_xyz(data: CameraData, bounds: Box, target: Point3) -> Point3:
    """
    Ensure that the target point is within the boundary Box along the x, y, and z-axis

    Assumes bounds.left < bounds.right, bounds.bottom < bounds.top, and bounds.near < bounds.right
    """
    x, y, z = data.position
    t_x, t_y, t_z = target
    dx = t_x - x
    dy = t_y - y
    dz = t_z - z
    if bounds.point_in_box((dx, dy, dz)):
        return data.position

    n_x = t_x - clamp(dx, bounds.left, bounds.right)
    n_y = t_y - clamp(dy, bounds.bottom, bounds.top)
    n_z = t_z - clamp(dz, bounds.near, bounds.far)

    return n_x, n_y, n_z



"""
Sprite Collect Coins with Different Levels

Simple program to show basic sprite usage.

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_collect_coins_diff_levels
"""

import random
import arcade

SPRITE_SCALING = 1.0

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sprite Collect Coins with Different Levels Example"


class FallingCoin(arcade.Sprite):
    """ Simple sprite that falls down """

    def update(self):
        """ Move the coin """

        # Fall down
        self.center_y -= 2

        # Did we go off the screen? If so, pop back to the top.
        if self.top < 0:
            self.bottom = WINDOW_HEIGHT


class RisingCoin(arcade.Sprite):
    """ Simple sprite that falls up """

    def update(self):
        """ Move the coin """

        # Move up
        self.center_y += 2

        # Did we go off the screen? If so, pop back to the bottom.
        if self.bottom > WINDOW_HEIGHT:
            self.top = 0


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        """ Initialize """

        # Call the parent class initializer
        super().__init__()

        # Variables that will hold sprite lists
        self.player_list = arcade.SpriteList()
        self.coin_list = arcade.SpriteList()

        # Set up the player info
        # Set up the player
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=SPRITE_SCALING,
        )
        self.player_list.append(self.player_sprite)

        self.score = 0
        self.level = 1

        # Don't show the mouse cursor
        self.window.set_mouse_visible(False)

        # Set the background color
        self.background_color = arcade.color.AMAZON

    def level_1(self):
        for i in range(20):

            # Create the coin instance
            coin = arcade.Sprite(
                ":resources:images/items/coinGold.png",
                scale=SPRITE_SCALING / 3,
            )

            # Position the coin
            coin.center_x = random.randrange(WINDOW_WIDTH)
            coin.center_y = random.randrange(WINDOW_HEIGHT)

            # Add the coin to the lists
            self.coin_list.append(coin)

    def level_2(self):
        for i in range(30):

            # Create the coin instance
            coin = FallingCoin(
                ":resources:images/items/coinBronze.png",
                scale=SPRITE_SCALING / 2,
            )

            # Position the coin
            coin.center_x = random.randrange(WINDOW_WIDTH)
            coin.center_y = random.randrange(WINDOW_HEIGHT, WINDOW_HEIGHT * 2)

            # Add the coin to the lists
            self.coin_list.append(coin)

    def level_3(self):
        for i in range(30):

            # Create the coin instance
            coin = RisingCoin(
                ":resources:images/items/coinSilver.png",
                scale=SPRITE_SCALING / 2,
            )

            # Position the coin
            coin.center_x = random.randrange(WINDOW_WIDTH)
            coin.center_y = random.randrange(-WINDOW_HEIGHT, 0)

            # Add the coin to the lists
            self.coin_list.append(coin)

    def reset(self):
        """ Set up the game and initialize the variables. """

        self.score = 0
        self.level = 1

        # Sprite lists
        self.coin_list.clear()

        self.player_sprite.center_x = 50
        self.player_sprite.center_y = 50

        self.level_1()

    def on_draw(self):
        """
        Render the screen.
        """

        # This command has to happen before we start drawing
        self.clear()

        # Draw all the sprites.
        arcade.draw_sprite(self.player_sprite)
        self.coin_list.draw()

        # Put the text on the screen.
        output = f"Score: {self.score}"
        arcade.draw_text(output, 10, 20, arcade.color.WHITE, 15)

        output = f"Level: {self.level}"
        arcade.draw_text(output, 10, 35, arcade.color.WHITE, 15)

    def on_mouse_motion(self, x, y, dx, dy):
        """
        Called whenever the mouse moves.
        """
        self.player_sprite.center_x = x
        self.player_sprite.center_y = y

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Call update on all sprites (The sprites don't do much in this
        # example though.)
        self.coin_list.update()

        # Generate a list of all sprites that collided with the player.
        hit_list = arcade.check_for_collision_with_list(self.player_sprite, self.coin_list)

        # Loop through each colliding sprite, remove it, and add to the score.
        for coin in hit_list:
            coin.remove_from_sprite_lists()
            self.score += 1

        # See if we should go to level 2
        if len(self.coin_list) == 0 and self.level == 1:
            self.level += 1
            self.level_2()
        # See if we should go to level 3
        elif len(self.coin_list) == 0 and self.level == 2:
            self.level += 1
            self.level_3()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.reset()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Example showing how er can emit particles with the gpu.
We work with transform shaders to do all the logic.

This example was created on a Raspberry PI 4.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.transform_emit
"""

# import struct
import random
from array import array
import arcade
from arcade.gl import BufferDescription

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Transform Emit - GPU"

class GameView(arcade.View):

    def __init__(self):
        super().__init__()
        self.ctx = self.window.ctx

        # Program to simply draw the points
        self.visualize_points_program = self.ctx.program(
            vertex_shader="""
                #version 330

                in vec2 in_pos;
                in vec3 in_col;

                out vec3 color;

                void main() {
                    gl_Position = vec4(in_pos, 0.0, 1.0);
                    color = in_col;
                }
                """,
            fragment_shader="""
                #version 330

                out vec4 fragColor;
                in vec3 color;

                void main() {
                    fragColor = vec4(color, 1.0);
                }
            """,
        )

        # Program to emit points into a buffer
        self.emit_program = self.ctx.program(
            vertex_shader="""
                #version 330
                #define M_PI 3.1415926535897932384626433832795
                uniform float time;

                out vec2 out_pos;
                out vec2 out_vel;

                float rand(float n) { return fract(sin(n) * 43758.5453123); }

                void main() {
                    float a = mod(time * float(gl_VertexID) + rand(time), M_PI * 2.0);
                    float r = clamp(rand(time + float(gl_VertexID)), 0.1, 0.9);

                    out_pos = vec2(0.0);
                    out_vel = vec2(sin(a), cos(a)) * r;
                }
            """,
            varyings=["out_pos", "out_vel"],
            varyings_capture_mode="separate",
        )

        self.move_program = self.ctx.program(
            vertex_shader="""
                #version 330

                in vec2 in_pos;
                in vec2 in_vel;

                out vec2 v_pos;
                out vec2 v_vel;

                void main() {
                    v_pos = in_pos;
                    v_vel = in_vel;
                }
            """,
            geometry_shader="""
                #version 330

                layout(points) in;
                layout(points, max_vertices = 1) out;

                uniform float dt;

                in vec2 v_pos[];
                in vec2 v_vel[];

                out vec2 out_pos;
                out vec2 out_vel;

                void main() {
                    vec2 pos = v_pos[0];
                    vec2 vel = v_vel[0];

                    if (pos.y > -1.0) {
                        out_pos = pos + vel * dt;
                        out_vel = vel - vec2(0.0, dt);
                        EmitVertex();
                    }

                }
            """,
            varyings=["out_pos", "out_vel"],
            varyings_capture_mode="separate",
        )

        # Configuration
        self.num_points = 10_000
        self.active_points = 0
        self.emit_max_points = 100

        # First set of positions and velocities
        self.buffer_pos_1 = self.ctx.buffer(reserve=self.num_points * 8)
        self.buffer_vel_1 = self.ctx.buffer(reserve=self.num_points * 8)

        # Second set of positions and velocities
        self.buffer_pos_2 = self.ctx.buffer(reserve=self.num_points * 8)
        self.buffer_vel_2 = self.ctx.buffer(reserve=self.num_points * 8)

        self.buffer_colors = self.ctx.buffer(
            data=array("f", [random.random() for _ in range(self.num_points * 3)])
        )

        # Geometry definition for drawing the two sets of positions
        self.draw_geometry_1 = self.ctx.geometry(
            [
                BufferDescription(self.buffer_pos_1, "2f", ("in_pos",)),
                BufferDescription(self.buffer_colors, "3f", ("in_col",)),
            ],
            mode=self.ctx.POINTS,
        )
        self.draw_geometry_2 = self.ctx.geometry(
            [
                BufferDescription(self.buffer_pos_2, "2f", ("in_pos",)),
                BufferDescription(self.buffer_colors, "3f", ("in_col",)),
            ],
            mode=self.ctx.POINTS,
        )

        # Geometry we use to move he points
        self.move_geometry_1 = self.ctx.geometry(
            [
                BufferDescription(self.buffer_pos_1, "2f", ("in_pos",)),
                BufferDescription(self.buffer_vel_1, "2f", ("in_vel",)),
            ],
            mode=self.ctx.POINTS,
        )
        self.move_geometry_2 = self.ctx.geometry(
            [
                BufferDescription(self.buffer_pos_2, "2f", ("in_pos",)),
                BufferDescription(self.buffer_vel_2, "2f", ("in_vel",)),
            ],
            mode=self.ctx.POINTS,
        )

        self.emit_geometry = self.ctx.geometry()
        self.query = self.ctx.query(primitives=True)

    def on_draw(self):
        self.clear()
        self.ctx.enable_only()

        self.draw_geometry_2.render(self.visualize_points_program)

        # Swap things around for the next frame
        self.draw_geometry_1, self.draw_geometry_2 = self.draw_geometry_2, self.draw_geometry_1
        self.buffer_pos_1, self.buffer_pos_2 = self.buffer_pos_2, self.buffer_pos_1
        self.buffer_vel_1, self.buffer_vel_2 = self.buffer_vel_2, self.buffer_vel_1

    def on_update(self, delta_time):
        # print("active points", self.active_points)

        # Do we have points to emit?
        if self.active_points < self.num_points:

            emit_count = min(self.num_points - self.active_points, self.emit_max_points)
            # print("Emitting", emit_count)

            # Emit some particles
            self.emit_program["time"] = self.window.time
            self.emit_geometry.transform(
                self.emit_program,
                [self.buffer_pos_1, self.buffer_vel_1],
                vertices=emit_count,
                buffer_offset=self.active_points * 8,
            )
            self.active_points += emit_count

        # Move the points. This will also purge dead particles
        self.move_program["dt"] = delta_time
        with self.query:
            self.move_geometry_1.transform(
                self.move_program,
                [self.buffer_pos_2, self.buffer_vel_2],
                vertices=self.active_points,
            )
        self.active_points = self.query.primitives_generated
        # print("after moving", self.active_points)


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Example showing how handle screen resizing.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.resizable_window
"""
import arcade

WINDOW_WIDTH = 500
WINDOW_HEIGHT = 500
WINDOW_TITLE = "Resizing Window Example"
START = 0
END = 2000
STEP = 50


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        super().__init__()

        self.background_color = arcade.color.WHITE

    def on_resize(self, width, height):
        """ This method is automatically called when the window is resized. """

        # Call the parent. Failing to do this will mess up the coordinates,
        # and default to 0,0 at the center and the edges being -1 to 1.
        super().on_resize(width, height)

        print(f"Window resized to: {width}, {height}")

    def on_draw(self):
        """ Render the screen. """

        self.clear()

        # Draw the y labels
        i = 0
        for y in range(START, END, STEP):
            arcade.draw_point(0, y, arcade.color.BLUE, 5)
            arcade.draw_text(f"{y}",
                             5, y,
                             arcade.color.BLACK,
                             12,
                             anchor_x="left", anchor_y="bottom")
            i += 1

        # Draw the x labels.
        i = 1
        for x in range(START + STEP, END, STEP):
            arcade.draw_point(x, 0, arcade.color.BLUE, 5)
            arcade.draw_text(f"{x}", x, 5, arcade.color.BLACK, 12,
                             anchor_x="left", anchor_y="bottom")
            i += 1


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE, resizable=True)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Gaussian kernel calculator
Adapted from:
https://observablehq.com/@jobleonard/gaussian-kernel-calculater
"""

import math

SQRT2 = math.sqrt(2)


def erf(x: float):
    a1 = 0.254829592
    a2 = -0.284496736
    a3 = 1.421413741
    a4 = -1.453152027
    a5 = 1.061405429
    p = 0.3275911

    t = 1.0 / (1.0 + p * abs(x))
    y = 1.0 - (((((a5 * t + a4) * t) + a3) * t + a2) * t + a1) * t * math.exp(-x * x)

    return math.copysign(1, x) * y


def def_int_gaussian(x: float, mu: float, sigma: float):
    return 0.5 * erf((x - mu) / (SQRT2 * sigma))


def gaussian_kernel(kernel_size: int = 5, sigma: float = 1, mu: float = 0, step: int = 1):
    end = 0.5 * kernel_size
    start = -end
    coeff = []
    sum = 0
    x = start
    last_int = def_int_gaussian(x, mu, sigma)
    while x < end:
        x += step
        new_int = def_int_gaussian(x, mu, sigma)
        c = new_int - last_int
        coeff.append(c)
        sum += c
        last_int = new_int

    # normalize
    sum_norm = 1 / sum
    for i in range(len(coeff)):
        coeff[i] *= sum_norm

    return coeff



from arcade.draw.rect import draw_texture_rect
from arcade.texture.texture import Texture
from arcade.types import LBWH, AnchorPoint
from arcade.window_commands import get_window


class StartFinishRenderData:
    """
    State data for offscreen rendering with :py:meth:`arcade.start_render` and
    :py:meth:`arcade.finish_render`. This is only meant for simple module level
    drawing like creating a static image we display repeatedly once the module
    has executed.

    Example::

        import arcade
        arcade.open_window(500, 500, "Picture")
        arcade.set_background_color(arcade.color.WHITE)
        # This renderer is permanently enabled here
        arcade.start_render()
        arcade.draw_text("Hello World", 190, 50, arcade.color.BLACK, 20)
        arcade.draw_circle_filled(250, 250, 100, arcade.color.RED)
        arcade.finish_render()
        # Repeatedly display the image produced between start and finish render
        arcade.run()

    This renderer is enabled by calling :py:meth:`arcade.start_render`. It's
    an irreversible action.

    Args:
        pixelated: Should the image be pixelated or smooth when scaled?
        blend: Should we draw with alpha blending enabled?
    """

    def __init__(self, pixelated: bool = False, blend: bool = True):
        from arcade.texture_atlas.atlas_default import DefaultTextureAtlas

        self.window = get_window()
        self.pixelated = pixelated
        self.blend = blend
        self.atlas = DefaultTextureAtlas(self.window.get_framebuffer_size(), border=0)
        self.texture = Texture.create_empty(
            "start_finish_render_texture", size=self.window.get_framebuffer_size()
        )
        self.atlas.add(self.texture)
        self._generator_func = None
        self.completed = False

    def begin(self):
        """
        Enable rendering into the buffer.

        Should only be called once followed by a call to :py:meth:`end`.
        """
        self.generator_func = self.atlas.render_into(
            self.texture,
            projection=(0, self.window.width, 0, self.window.height),
        )
        fbo = self.generator_func.__enter__()
        fbo.clear(color=self.window.background_color)

        if self.blend:
            self.window.ctx.enable(self.window.ctx.BLEND)

    def end(self):
        """Switch back to rendering into the window"""
        self.generator_func.__exit__(None, None, None)
        self.completed = True

    def draw(self):
        """
        Draw the buffer to the screen attempting to preserve the aspect ratio.
        """
        # Stretch the texture to the window size with black bars if needed
        w, h = self.window.get_size()
        min_factor = min(w / self.texture.width, h / self.texture.height)
        region = LBWH(0, 0, self.texture.width, self.texture.height).scale(
            min_factor, anchor=AnchorPoint.BOTTOM_LEFT
        )
        region = region.move(dx=(w - region.width) / 2, dy=(h - region.height) / 2)

        draw_texture_rect(
            self.texture, region, pixelated=self.pixelated, blend=False, atlas=self.atlas
        )



from typing import Any
from .hit_box import HitBoxCache
from .texture import TextureCache
from .image_data import ImageDataCache


def crate_str_from_values(*args, sep: str = "_") -> str:
    """
    Create a string from a list of parameters.

    Example::

        >> entries = ["blue", 5]
        >> crate_str_from_list("blue", 5)
        "blue_5"

    Args:
        params: List of parameters to create a string from.
        sep: Separator to use between parameters.
    """
    return sep.join([str(x) for x in args])


def crate_str_from_list(entries: list[Any], sep: str = "_") -> str:
    """
    Create a string from a list of parameters.

    Example::

        >> entries = ["blue", 5]
        >> crate_str_from_list(entries)
        "blue_5"

    Args:
        entries: List of parameters to create a string from.
        sep: Separator to use between parameters.
    """
    return crate_str_from_values(*entries, sep=sep)


__all__ = [
    "HitBoxCache",
    "TextureCache",
    "crate_str_from_list",
    "crate_str_from_values",
    "ImageDataCache",
]
